import os

import numpy as np
import pandas as pd
import pyarrow as pa
import pytest
from hats import pixel_math
from hats.io import paths
from hats.pixel_math.healpix_pixel import HealpixPixel

from hats_import.margin_cache import margin_cache_map_reduce
from hats_import.pipeline_resume_plan import get_pixel_cache_directory

keep_cols = ["weird_ra", "weird_dec"]

drop_cols = [
    "partition_order",
    "partition_pixel",
    "margin_check",
    "margin_pixel",
    "is_trunc",
]


def validate_result_dataframe(df_path, expected_len):
    res_df = pd.read_parquet(df_path)

    assert len(res_df) == expected_len

    cols = res_df.columns.values.tolist()

    for col in keep_cols:
        assert col in cols

    for col in drop_cols:
        assert col not in cols


@pytest.mark.timeout(5)
def test_to_pixel_shard_equator(tmp_path, basic_data_shard_df):
    margin_cache_map_reduce._to_pixel_shard(
        pa.Table.from_pandas(basic_data_shard_df),
        pixel=HealpixPixel(1, 21),
        output_path=tmp_path,
        source_pixel=HealpixPixel(1, 0),
        healpix_column="_healpix_29",
    )

    path = tmp_path / "order_1" / "dir_0" / "pixel_21" / "dataset" / "Norder=1" / "Dir=0" / "Npix=0.parquet"

    assert os.path.exists(path)

    validate_result_dataframe(path, 360)


@pytest.mark.timeout(5)
def test_to_pixel_shard_polar(tmp_path, polar_data_shard_df):
    margin_cache_map_reduce._to_pixel_shard(
        pa.Table.from_pandas(polar_data_shard_df),
        pixel=HealpixPixel(2, 15),
        output_path=tmp_path,
        source_pixel=HealpixPixel(2, 0),
        healpix_column="_healpix_29",
    )

    path = tmp_path / "order_2" / "dir_0" / "pixel_15" / "dataset" / "Norder=2" / "Dir=0" / "Npix=0.parquet"

    assert os.path.exists(path)

    validate_result_dataframe(path, 360)


def test_map_pixel_shards_error(tmp_path, capsys):
    """Test error behavior on reduce stage. e.g. by not creating the original
    catalog parquet files."""
    with pytest.raises(FileNotFoundError):
        margin_cache_map_reduce.map_pixel_shards(
            paths.pixel_catalog_file(tmp_path, HealpixPixel(1, 0)),
            mapping_key="1_21",
            source_pixel=HealpixPixel(1, 0),
            original_catalog_metadata="",
            margin_pair_file="",
            output_path=tmp_path,
            margin_order=4,
            healpix_column="_healpix_29",
            healpix_order=29,
        )

    captured = capsys.readouterr()
    assert "Parquet file does not exist" in captured.out


@pytest.mark.timeout(15)
def test_map_pixel_shards_coarse(tmp_path, test_data_dir, small_sky_source_catalog):
    """Test basic mapping behavior, without fine filtering enabled."""
    intermediate_dir = tmp_path / "intermediate"
    os.makedirs(intermediate_dir / "mapping")
    margin_cache_map_reduce.map_pixel_shards(
        paths.pixel_catalog_file(small_sky_source_catalog, HealpixPixel(1, 47)),
        source_pixel=HealpixPixel(1, 47),
        mapping_key="1_47",
        original_catalog_metadata=small_sky_source_catalog / "dataset" / "_common_metadata",
        margin_pair_file=test_data_dir / "margin_pairs" / "small_sky_source_pairs.csv",
        output_path=intermediate_dir,
        margin_order=3,
        healpix_column="_healpix_29",
        healpix_order=29,
    )

    path = (
        intermediate_dir
        / "order_2"
        / "dir_0"
        / "pixel_182"
        / "dataset"
        / "Norder=1"
        / "Dir=0"
        / "Npix=47.parquet"
    )
    assert os.path.exists(path)
    res_df = pd.read_parquet(path)
    assert len(res_df) == 1386

    path = (
        intermediate_dir
        / "order_2"
        / "dir_0"
        / "pixel_185"
        / "dataset"
        / "Norder=1"
        / "Dir=0"
        / "Npix=47.parquet"
    )
    assert os.path.exists(path)
    res_df = pd.read_parquet(path)
    assert len(res_df) == 1978


def test_reduce_margin_shards(tmp_path):
    intermediate_dir = tmp_path / "intermediate"
    partition_dir = get_pixel_cache_directory(intermediate_dir, HealpixPixel(1, 21))
    shard_dir = paths.pixel_directory(partition_dir, 1, 21)

    os.makedirs(shard_dir)
    os.makedirs(intermediate_dir / "reducing")

    first_shard_path = paths.pixel_catalog_file(partition_dir, HealpixPixel(1, 0))
    second_shard_path = paths.pixel_catalog_file(partition_dir, HealpixPixel(1, 1))

    ras = np.arange(0.0, 360.0)
    dec = np.full(360, 0.0)
    hats_indexes = pixel_math.compute_spatial_index(ras, dec)

    basic_data_shard_df = pd.DataFrame(
        data=zip(hats_indexes, ras, dec),
        columns=[
            "_healpix_29",
            "weird_ra",
            "weird_dec",
        ],
    )

    basic_data_shard_df.to_parquet(first_shard_path)
    basic_data_shard_df.to_parquet(second_shard_path)

    margin_cache_map_reduce.reduce_margin_shards(
        intermediate_dir,
        "1_21",
        tmp_path,
        1,
        21,
        delete_intermediate_parquet_files=False,
    )

    result_path = paths.pixel_catalog_file(tmp_path, HealpixPixel(1, 21))

    validate_result_dataframe(result_path, 720)
    assert os.path.exists(shard_dir)

    # Run again with delete_intermediate_parquet_files. shard_dir doesn't exist at the end.
    margin_cache_map_reduce.reduce_margin_shards(
        intermediate_dir,
        "1_21",
        tmp_path,
        1,
        21,
        delete_intermediate_parquet_files=True,
    )

    result_path = paths.pixel_catalog_file(tmp_path, HealpixPixel(1, 21))

    validate_result_dataframe(result_path, 720)
    assert not os.path.exists(shard_dir)



"""Utility to hold all arguments required throughout indexing"""

from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path

from hats import read_hats
from hats.catalog import Catalog, TableProperties
from hats.io.validation import is_valid_catalog
from upath import UPath

from hats_import.runtime_arguments import RuntimeArguments


@dataclass
class IndexArguments(RuntimeArguments):
    """Data class for holding indexing arguments"""

    ## Input
    input_catalog_path: str | Path | UPath | None = None
    input_catalog: Catalog | None = None
    indexing_column: str = ""
    extra_columns: list[str] = field(default_factory=list)

    ## Output
    include_healpix_29: bool = True
    """Include the healpix-based hats spatial index."""
    include_order_pixel: bool = True
    """Include partitioning columns, Norder, Dir, and Npix. You probably want to keep these!"""
    include_radec: bool = False
    """Include the ra/dec coordinates of the row."""
    drop_duplicates: bool = True
    """Should we check for duplicate rows (including new indexing column),
    and remove duplicates before writing to new index catalog?
    If you know that your data will not have duplicates (e.g. an index over
    a unique primary key), set to False to avoid unnecessary work."""

    ## Compute parameters
    compute_partition_size: int = 1_000_000_000
    """partition size used when creating leaf parquet files."""
    division_hints: list | None = None
    """Hints used when splitting up the rows by the new index. If you have
    some prior knowledge about the distribution of your indexing_column, 
    providing it here can speed up calculations dramatically. Note that
    these will NOT necessarily be the divisions that the data is partitioned
    along."""

    def __post_init__(self):
        self._check_arguments()

    def _check_arguments(self):
        super()._check_arguments()
        if not self.input_catalog_path:
            raise ValueError("input_catalog_path is required")
        if not self.indexing_column:
            raise ValueError("indexing_column is required")

        if not self.include_healpix_29 and not self.include_order_pixel:
            raise ValueError("At least one of include_healpix_29 or include_order_pixel must be True")

        if not is_valid_catalog(self.input_catalog_path):
            raise ValueError("input_catalog_path not a valid catalog")
        self.input_catalog = read_hats(catalog_path=self.input_catalog_path)
        if self.include_radec:
            catalog_info = self.input_catalog.catalog_info
            self.extra_columns.extend([catalog_info.ra_column, catalog_info.dec_column])
        if len(self.extra_columns) > 0:
            # check that they're in the schema
            schema = self.input_catalog.schema
            missing_fields = [x for x in self.extra_columns if schema.get_field_index(x) == -1]
            if len(missing_fields):
                raise ValueError(f"Some requested columns not in input catalog ({','.join(missing_fields)})")
        # Remove duplicates, preserving order
        extra_columns = []
        for x in self.extra_columns:
            if x not in extra_columns:
                extra_columns.append(x)
        self.extra_columns = extra_columns

        if self.compute_partition_size < 100_000:
            raise ValueError("compute_partition_size must be at least 100_000")

    def to_table_properties(self, total_rows: int) -> TableProperties:
        """Catalog-type-specific dataset info."""
        info = {
            "catalog_name": self.output_artifact_name,
            "catalog_type": "index",
            "total_rows": total_rows,
            "primary_catalog": str(self.input_catalog_path),
            "indexing_column": self.indexing_column,
            "extra_columns": self.extra_columns,
        } | self.extra_property_dict()

        return TableProperties(**info)



import pytest

from hats_import.collection.arguments import CollectionArguments, _pretty_print_angle


def test_missing_required(tmp_path):
    """No arguments provided. Should error for required args."""
    CollectionArguments(
        output_artifact_name="good_name",
        output_path=tmp_path,
    )

    with pytest.raises(ValueError, match="invalid character"):
        CollectionArguments(
            output_artifact_name="bad_a$$_name",
            output_path=tmp_path,
        )
    with pytest.raises(ValueError):
        CollectionArguments()

    ## Output path is missing
    with pytest.raises(ValueError, match="output_path"):
        CollectionArguments(
            output_artifact_name="catalog",
            output_path="",
        )

    ## Output catalog name is missing
    with pytest.raises(ValueError, match="output_artifact_name"):
        CollectionArguments(
            output_artifact_name="",
            output_path=tmp_path,
        )


def test_subcatalog_table_properties(tmp_path, blank_data_dir):
    args = (
        CollectionArguments(
            output_artifact_name="good_name",
            output_path=tmp_path,
            progress_bar=False,
            addl_hats_properties={"hats_cols_default": "id, mjd", "obs_regime": "Optical"},
        )
        .catalog(
            input_path=blank_data_dir,
            file_reader="csv",
        )
        .add_margin(margin_threshold=5.0)
        .add_margin(margin_threshold=15.0)
    )

    catalog_info = args.catalog_args.to_table_properties(
        total_rows=10, highest_order=4, moc_sky_fraction=22 / 7
    )

    assert catalog_info.catalog_name == "good_name"
    assert catalog_info.total_rows == 10
    assert catalog_info.default_columns == ["id", "mjd"]
    assert catalog_info.__pydantic_extra__["obs_regime"] == "Optical"


def test_subcatalog_wrong_order(tmp_path, blank_data_dir):
    with pytest.raises(ValueError, match="add catalog arguments"):
        CollectionArguments(
            output_artifact_name="good_name",
            output_path=tmp_path,
            progress_bar=False,
            addl_hats_properties={"hats_cols_default": "id, mjd", "obs_regime": "Optical"},
        ).add_margin(margin_threshold=5.0).add_margin(margin_threshold=15.0)

    with pytest.raises(ValueError, match="add catalog arguments"):
        CollectionArguments(
            output_artifact_name="good_name",
            output_path=tmp_path,
            progress_bar=False,
            addl_hats_properties={"hats_cols_default": "id, mjd", "obs_regime": "Optical"},
        ).add_index(indexing_column="id")

    with pytest.raises(ValueError, match="exactly once"):
        CollectionArguments(
            output_artifact_name="good_name",
            output_path=tmp_path,
            progress_bar=False,
            addl_hats_properties={"hats_cols_default": "id, mjd", "obs_regime": "Optical"},
        ).catalog(
            input_path=blank_data_dir,
            file_reader="csv",
        ).catalog(
            input_path=blank_data_dir,
            file_reader="csv",
        )

    args = CollectionArguments(
        output_artifact_name="good_name",
        output_path=tmp_path,
        progress_bar=False,
        addl_hats_properties={"hats_cols_default": "id, mjd", "obs_regime": "Optical"},
    )
    with pytest.raises(ValueError, match="add catalog arguments"):
        args.get_catalog_args()
    with pytest.raises(ValueError, match="add catalog arguments"):
        args.get_margin_args()
    with pytest.raises(ValueError, match="add catalog arguments"):
        args.get_index_args()
    with pytest.raises(ValueError, match="add catalog arguments"):
        args.to_collection_properties()


def test_subcatalog_existing_catalog(tmp_path, small_sky_object_catalog):
    args = (
        CollectionArguments(
            output_artifact_name="small_sky_collection",
            output_path=tmp_path,
            progress_bar=False,
            tmp_dir=tmp_path,
        )
        .catalog(
            catalog_path=small_sky_object_catalog,
        )
        .add_margin(margin_threshold=5.0)
        .add_margin(margin_threshold=15.0)
        .add_index(indexing_column="id")
    )

    margin_args = args.get_margin_args()
    assert len(margin_args) == 2

    # test idempotency
    margin_args = args.get_margin_args()
    assert len(margin_args) == 2

    index_args = args.get_index_args()
    assert len(index_args) == 1

    # test idempotency
    index_args = args.get_index_args()
    assert len(index_args) == 1


def test_index_bad_values(tmp_path, small_sky_object_catalog):
    with pytest.raises(ValueError, match="indexing_column is required"):
        (
            CollectionArguments(
                output_artifact_name="small_sky_collection",
                output_path=tmp_path,
                progress_bar=False,
                tmp_dir=tmp_path,
            )
            .catalog(
                catalog_path=small_sky_object_catalog,
            )
            .add_index()
        )

    args = (
        CollectionArguments(
            output_artifact_name="small_sky_collection",
            output_path=tmp_path,
            progress_bar=False,
            tmp_dir=tmp_path,
        )
        .catalog(
            catalog_path=small_sky_object_catalog,
        )
        .add_index(indexing_column="id", extra_columns=["not", "there"])
    )

    with pytest.raises(ValueError, match="not in input catalog"):
        args.get_index_args()


def test_margin_bad_values(tmp_path, small_sky_object_catalog):
    with pytest.raises(ValueError, match="threshold required"):
        (
            CollectionArguments(
                output_artifact_name="small_sky_collection",
                output_path=tmp_path,
                progress_bar=False,
                tmp_dir=tmp_path,
            )
            .catalog(
                catalog_path=small_sky_object_catalog,
            )
            .add_margin()
        )

    args = (
        CollectionArguments(
            output_artifact_name="small_sky_collection",
            output_path=tmp_path,
            progress_bar=False,
            tmp_dir=tmp_path,
        )
        .catalog(
            catalog_path=small_sky_object_catalog,
        )
        .add_margin(margin_threshold=1_000_000)
    )

    with pytest.raises(ValueError, match="higher order"):
        args.get_margin_args()


def test_collection_properties_basic(tmp_path, blank_data_dir):
    args = CollectionArguments(
        output_artifact_name="good_name",
        output_path=tmp_path,
        progress_bar=False,
        addl_hats_properties={"obs_regime": "Optical"},
    ).catalog(
        input_path=blank_data_dir,
        file_reader="csv",
        addl_hats_properties={"hats_cols_default": "id, mjd"},
    )

    collection_info = args.to_collection_properties()
    assert collection_info.name == "good_name"
    assert collection_info.hats_primary_table_url == "good_name"
    assert collection_info.all_margins is None
    assert collection_info.all_indexes is None
    assert collection_info.__pydantic_extra__["obs_regime"] == "Optical"


def test_collection_properties_supplemental(tmp_path, small_sky_object_catalog):
    args = (
        CollectionArguments(
            output_artifact_name="small_sky_collection",
            new_catalog_name="small_sky",
            output_path=tmp_path,
            progress_bar=False,
            addl_hats_properties={"obs_regime": "Optical"},
        )
        .catalog(
            catalog_path=small_sky_object_catalog,
        )
        .add_margin(margin_threshold=5.0, is_default=True)
        .add_margin(margin_threshold=35.0)
        .add_index(indexing_column="id")
    )

    collection_info = args.to_collection_properties()
    assert collection_info.name == "small_sky_collection"
    assert len(collection_info.all_margins) == 2
    assert collection_info.default_margin == "small_sky_object_catalog_5arcs"
    assert len(collection_info.all_indexes) == 1
    assert collection_info.__pydantic_extra__["obs_regime"] == "Optical"

    margin_args = args.get_margin_args()[0]
    assert margin_args.output_artifact_name == "small_sky_object_catalog_5arcs"

    margin_args = args.get_margin_args()[1]
    assert margin_args.output_artifact_name == "small_sky_object_catalog_35arcs"

    index_args = args.get_index_args()[0]
    assert index_args.output_artifact_name == "small_sky_object_catalog_id"


def test_collection_properties_existing_supplemental(tmp_path, small_sky_object_catalog):
    """If we have already created a catalog or supplemental table in the
    collection, confirm that the collection properties are written with
    the union of supplemental tables."""
    args = (
        CollectionArguments(
            output_artifact_name="small_sky_collection",
            new_catalog_name="small_sky",
            output_path=tmp_path,
            progress_bar=False,
            addl_hats_properties={"obs_regime": "Optical"},
        ).catalog(
            catalog_path=small_sky_object_catalog,
        )
        # .add_margin(margin_threshold=5.0, is_default=True)
        # .add_margin(margin_threshold=35.0)
        # .add_index(indexing_column="id")
    )

    collection_info = args.to_collection_properties()
    collection_info.to_properties_file(args.catalog_path)

    args = (
        CollectionArguments(
            output_artifact_name="small_sky_collection",
            new_catalog_name="small_sky",
            output_path=tmp_path,
            progress_bar=False,
        )
        .catalog(
            catalog_path=small_sky_object_catalog,
        )
        .add_margin(margin_threshold=5.0, is_default=True)
        # .add_margin(margin_threshold=35.0)
        .add_index(indexing_column="id")
    )

    collection_info = args.to_collection_properties()
    collection_info.to_properties_file(args.catalog_path)

    args = (
        CollectionArguments(
            output_artifact_name="small_sky_collection",
            new_catalog_name="small_sky",
            output_path=tmp_path,
            progress_bar=False,
        ).catalog(
            catalog_path=small_sky_object_catalog,
        )
        # .add_margin(margin_threshold=5.0, is_default=True)
        .add_margin(margin_threshold=35.0)
        # .add_index(indexing_column="id")
    )

    collection_info = args.to_collection_properties()
    assert collection_info.name == "small_sky_collection"
    assert len(collection_info.all_margins) == 2
    assert collection_info.default_margin == "small_sky_object_catalog_5arcs"
    assert len(collection_info.all_indexes) == 1
    assert collection_info.__pydantic_extra__["obs_regime"] == "Optical"


def test_collection_default_margin(tmp_path, small_sky_object_catalog):
    """Test behavior of the default margin setting."""
    ## Easy case - name derived from primary catalog
    args = (
        CollectionArguments(
            output_artifact_name="small_sky_collection",
            new_catalog_name="small_sky",
            output_path=tmp_path,
            progress_bar=False,
        )
        .catalog(
            catalog_path=small_sky_object_catalog,
        )
        .add_margin(margin_threshold=5.0, is_default=True)
        .add_margin(margin_threshold=35.0)
    )

    collection_info = args.to_collection_properties()
    assert collection_info.name == "small_sky_collection"
    assert collection_info.all_margins == [
        "small_sky_object_catalog_5arcs",
        "small_sky_object_catalog_35arcs",
    ]
    assert collection_info.default_margin == "small_sky_object_catalog_5arcs"

    ## Margin name passed explicitly
    args = (
        CollectionArguments(
            output_artifact_name="small_sky_collection",
            new_catalog_name="small_sky",
            output_path=tmp_path,
            progress_bar=False,
        )
        .catalog(
            catalog_path=small_sky_object_catalog,
        )
        .add_margin(output_artifact_name="my_5arcs_margin", margin_threshold=5.0, is_default=True)
        .add_margin(margin_threshold=35.0)
    )

    collection_info = args.to_collection_properties()
    assert collection_info.name == "small_sky_collection"
    assert collection_info.all_margins == ["my_5arcs_margin", "small_sky_object_catalog_35arcs"]
    assert collection_info.default_margin == "my_5arcs_margin"

    with pytest.raises(ValueError, match="Only one margin catalog may be the default"):
        ## Only one default margin allowed.
        (
            CollectionArguments(
                output_artifact_name="small_sky_collection",
                new_catalog_name="small_sky",
                output_path=tmp_path,
                progress_bar=False,
            )
            .catalog(
                catalog_path=small_sky_object_catalog,
            )
            .add_margin(margin_threshold=5.0, is_default=True)
            .add_margin(margin_threshold=35.0, is_default=True)
        )


def test_pretty_print_angle():
    assert _pretty_print_angle(1) == "1arcs"
    assert _pretty_print_angle(100) == "1arcmin"
    assert _pretty_print_angle(10_000) == "2deg"
    assert _pretty_print_angle(0.1) == "100msec"



"""Run pass/fail checks and generate verification report of existing hats table."""

import datetime
import re
from dataclasses import dataclass, field
from time import perf_counter
from urllib.parse import unquote

import hats
import hats.io.paths
import hats.io.validation
import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.dataset as pds
from hats import read_hats
from hats.catalog.catalog_collection import CatalogCollection
from hats.pixel_math.spatial_index import SPATIAL_INDEX_COLUMN

import hats_import
from hats_import.verification.arguments import VerificationArguments


# pylint: disable=too-many-lines
def run(args: VerificationArguments) -> "Verifier":
    """Create a `Verifier` using `args`, run all tests, and write a verification report.

    Parameters
    ----------
    args : VerificationArguments
        Arguments to construct the Verifier.

    Returns
    -------
    Verifier
        The `Verifier` instance used to perform the tests. The `results_df` property contains
        the same information as written to the output report.

    Raises
    ------
    TypeError
        If `args` is not provided or is not an instance of `VerificationArguments`.
    """
    if not args:
        raise TypeError("args is required and should be type VerificationArguments")
    if not isinstance(args, VerificationArguments):
        raise TypeError("args must be type VerificationArguments")

    start = perf_counter()
    verifier = Verifier.from_args(args)
    verifier.run()

    if args.verbose:
        print(f"Elapsed time (seconds): {perf_counter()-start:.2f}")

    return verifier


def now() -> str:
    """Get the current time as a string."""
    return datetime.datetime.now(datetime.timezone.utc).strftime("%Y/%m/%d %H:%M:%S %Z")


@dataclass(kw_only=True, frozen=True)
class Result:
    """Verification test result for a single test."""

    datetime: str = field(default_factory=now)
    """The date and time when the test was run."""
    passed: bool = field()
    """Whether the test passed."""
    test: str = field()
    """Test name."""
    target: str = field()
    """The file(s) targeted by the test."""
    description: str = field()
    """Test description."""
    bad_files: list[str] = field(default_factory=list)
    """List of additional files that caused the test to fail (empty if none or not applicable)."""


@dataclass(kw_only=True)
class Verifier:
    """Run verification tests. To create an instance of this class, use `Verifier.from_args`."""

    args: VerificationArguments = field()
    """Arguments to use during verification."""
    metadata_ds: pds.Dataset = field()
    """Pyarrow dataset, loaded from the _metadata file."""
    files_ds: pds.Dataset = field()
    """Pyarrow dataset, loaded from the parquet data files."""
    common_metadata_schema: pa.Schema = field()
    """Pyarrow schema, loaded from the _common_metadata file."""
    constructed_truth_schema: pa.Schema = field()
    """Pyarrow schema treated as truth during verification. This is constructed
    from `common_metadata_schema` and `args.truth_schema`. `common_metadata_schema`
    is used for hats-specific columns. If provided, `args.truth_schema` is used
    for all other columns plus metadata, otherwise `common_metadata_schema` is used.
    """
    results: list[Result] = field(default_factory=list)
    """List of results, one for each test that has been done."""

    @classmethod
    def from_args(cls, args: VerificationArguments) -> "Verifier":
        """Create a `Verifier` with initialized datasets and schemas based on `args`.

        Parameters
        ----------
        args : VerificationArguments
            Arguments for the Verifier.

        Returns
        -------
        Verifier
        """

        args.output_path.mkdir(exist_ok=True, parents=True)

        if args.verbose:
            print("Loading dataset and schema.")
        parquet_fs = args.input_catalog_path.fs

        ## Fetch all sub-URLs that could contain hats leaf files.
        all_files = []
        for child in args.input_dataset_path.rglob("Norder*/**/*"):
            if not child.is_dir():
                all_files.append(unquote(child.path))

        files_ds = pds.dataset(all_files, filesystem=parquet_fs)
        metadata_ds = pds.parquet_dataset(
            hats.io.paths.get_parquet_metadata_pointer(args.input_catalog_path), filesystem=parquet_fs
        )

        input_truth_schema = None
        if args.truth_schema is not None:
            input_truth_schema = pds.parquet_dataset(args.truth_schema, filesystem=parquet_fs).schema
        common_metadata_schema = pds.parquet_dataset(
            hats.io.paths.get_common_metadata_pointer(args.input_catalog_path), filesystem=parquet_fs
        ).schema
        constructed_truth_schema = cls._construct_truth_schema(
            input_truth_schema=input_truth_schema, common_metadata_schema=common_metadata_schema
        )

        return cls(
            args=args,
            metadata_ds=metadata_ds,
            files_ds=files_ds,
            common_metadata_schema=common_metadata_schema,
            constructed_truth_schema=constructed_truth_schema,
        )

    @property
    def results_df(self) -> pd.DataFrame:
        """Test results as a dataframe."""
        return pd.DataFrame(self.results)

    @property
    def all_tests_passed(self):
        """Simple pass/fail if all of the test results have passed."""
        return np.all([res.passed for res in self.results])

    def run(self) -> None:
        """Run all tests and write a verification report. See `results_df` property or
        written report for results.
        """
        self.test_is_valid_catalog()
        self.test_file_sets()
        self.test_num_rows()
        self.test_schemas()

        self.write_results()

    def test_is_valid_catalog(self) -> bool:
        """Test if the provided catalog is a valid HATS catalog. Add one `Result` to `results`.

        Returns
        -------
            bool: True if the test passed, else False.
        """
        test, description = "valid hats", "Test hats.io.validation.is_valid_catalog."
        target = self.args.input_catalog_path
        self.print_if_verbose(f"\nStarting: {description}")

        passed = hats.io.validation.is_valid_catalog(target, strict=True, verbose=self.args.verbose)
        self.results.append(Result(test=test, description=description, passed=passed, target=target.name))
        self.print_if_verbose(f"Result: {'PASSED' if passed else 'FAILED'}")
        return passed

    def test_file_sets(self) -> bool:
        """Test that files in _metadata match the parquet files on disk. Add one `Result` to `results`.

        This is a simple test that can be especially useful to run after copying or moving
        the catalog to a different local or cloud-based destination.

        Returns
        -------
            bool: True if the file sets match, else False.
        """
        # info for the report
        description = "Test that files in _metadata match the data files on disk."
        self.print_if_verbose(f"\nStarting: {description}")

        files_ds_files = self._relative_paths(self.files_ds.files)
        metadata_ds_files = self._relative_paths(self.metadata_ds.files)
        failed_files = list(set(files_ds_files).symmetric_difference(metadata_ds_files))
        passed = len(failed_files) == 0
        self.results.append(
            Result(
                passed=passed,
                test="file sets",
                target="_metadata vs data files",
                bad_files=failed_files,
                description=description,
            )
        )

        self.print_if_verbose(f"Result: {'PASSED' if passed else 'FAILED'}")
        return passed

    def test_num_rows(self) -> bool:
        """Test the number of rows in the dataset. Add `Result`s to `results`.

        Row counts in parquet file footers are compared with the '_metadata' file,
        HATS 'properties' file, and (if provided) the user-supplied truth.

        Returns
        -------
        bool: True if all checks pass, else False.
        """
        test = "num rows"
        description = "Test that number of rows are equal."
        self.print_if_verbose(f"\nStarting: {description}")

        catalog = read_hats(self.args.input_catalog_path)

        if isinstance(catalog, CatalogCollection):
            catalog = catalog.main_catalog
        catalog_prop_len = catalog.catalog_info.total_rows

        # get the number of rows in each file, indexed by file path. we treat this as truth.
        files_df = self._load_nrows(self.files_ds)
        files_df_sum = files_df.num_rows.sum()
        files_df_total = f"file footers ({files_df_sum:,})"

        target = "file footers vs catalog properties"
        self.print_if_verbose(f"\t{target}")
        passed_cat = catalog_prop_len == files_df_sum
        _description = f" {files_df_total} vs catalog properties ({catalog_prop_len:,})."
        self.results.append(
            Result(passed=passed_cat, test=test, target=target, description=description + _description)
        )

        # check _metadata
        target = "file footers vs _metadata"
        self.print_if_verbose(f"\t{target}")
        metadata_df = self._load_nrows(self.metadata_ds)
        row_diff = files_df - metadata_df
        failed_frags = row_diff.loc[row_diff.num_rows != 0].index.to_list()
        passed_md = len(failed_frags) == 0
        _description = f" {files_df_total} vs _metadata ({metadata_df.num_rows.sum():,})."
        self.results.append(
            Result(
                passed=passed_md,
                test=test,
                bad_files=failed_frags,
                target=target,
                description=description + _description,
            )
        )

        # check user-supplied total, if provided
        if self.args.truth_total_rows is not None:
            target = "file footers vs truth"
            self.print_if_verbose(f"\t{target}")
            passed_th = self.args.truth_total_rows == files_df_sum
            _description = f" {files_df_total} vs user-provided truth ({self.args.truth_total_rows:,})."
            self.results.append(
                Result(passed=passed_th, test=test, target=target, description=description + _description)
            )
        else:
            passed_th = True  # this test did not fail. this is only needed for the return value.

        all_passed = all([passed_md, passed_th, passed_cat])
        self.print_if_verbose(f"Result: {'PASSED' if all_passed else 'FAILED'}")
        return all_passed

    def _load_nrows(self, dataset: pds.Dataset) -> pd.DataFrame:
        """Load the number of rows in each file in the dataset.

        Parameters
        ----------
        dataset : pyarrow.dataset.Dataset
            The dataset from which to load the number of rows.

        Returns
        -------
        pd.DataFrame: A DataFrame with the number of rows per file, indexed by file path.
        """
        num_rows = [frag.metadata.num_rows for frag in dataset.get_fragments()]
        frag_names = self._relative_paths([frag.path for frag in dataset.get_fragments()])
        nrows_df = pd.DataFrame({"num_rows": num_rows, "frag_path": frag_names})
        nrows_df = nrows_df.set_index("frag_path").sort_index()
        return nrows_df

    def test_schemas(self) -> bool:
        """Test the equality of schemas. Add `Result`s to `results`.

        This performs three tests:
        1. `common_metadata_schema` vs `constructed_truth_schema`.
        2. `metadata_ds.schema` vs `constructed_truth_schema`.
        3. File footers vs `constructed_truth_schema`.

        Returns
        -------
        bool: True if all tests pass, else False.
        """
        # info for the report
        _include_md = "including metadata" if self.args.check_metadata else "excluding metadata"
        test_info = {"test": "schema", "description": f"Test that schemas are equal, {_include_md}."}
        self.print_if_verbose(f"\nStarting: {test_info['description']}")

        passed_cm = self._test_schema__common_metadata(test_info)
        passed_md = self._test_schema__metadata(test_info)
        passed_ff = self._test_schema_file_footers(test_info)

        all_passed = all([passed_cm, passed_md, passed_ff])
        self.print_if_verbose(f"Result: {'PASSED' if all_passed else 'FAILED'}")
        return all_passed

    @staticmethod
    def _construct_truth_schema(
        *, input_truth_schema: pa.Schema | None, common_metadata_schema: pa.Schema
    ) -> pa.Schema:
        """Copy of `input_truth_schema` with HATS fields added from `common_metadata_schema`.

        If `input_truth_schema` is not provided, this is just `common_metadata_schema`.

        Parameters
        ----------
        input_truth_schema : pyarrow.Schema or None
            The input truth schema, if provided.
        common_metadata_schema : pyarrow.Schema
            The common metadata schema.

        Returns
        -------
        pyarrow.Schema
            The constructed truth schema.
        """
        if input_truth_schema is None:
            return common_metadata_schema

        hats_cols = ["Norder", "Dir", "Npix"]
        hats_idx_fields = []
        if SPATIAL_INDEX_COLUMN in common_metadata_schema.names:
            hats_cols.append(SPATIAL_INDEX_COLUMN)
            hats_idx_fields.append(common_metadata_schema.field(SPATIAL_INDEX_COLUMN))
        input_truth_fields = [fld for fld in input_truth_schema if fld.name not in hats_cols]

        constructed_fields = hats_idx_fields + input_truth_fields
        constructed_schema = pa.schema(constructed_fields).with_metadata(input_truth_schema.metadata)
        return constructed_schema

    def _test_schema__common_metadata(self, test_info: dict) -> bool:
        """Test `common_metadata_schema` against `constructed_truth_schema`.

        Parameters
        ----------
        test_info : dict
            Information about this test for the reported results.

        Returns
        -------
        bool: True if all tests pass, else False.
        """
        targets = "_common_metadata vs truth"
        self.print_if_verbose(f"\t{targets}")
        passed = self.common_metadata_schema.equals(
            self.constructed_truth_schema, check_metadata=self.args.check_metadata
        )
        self.results.append(
            Result(
                passed=passed, target=targets, test=test_info["test"], description=test_info["description"]
            )
        )
        return passed

    def _test_schema__metadata(self, test_info: dict) -> bool:
        """Test _metadata schema against the truth schema.

        Parameters
        ----------
        test_info : dict
            Information about this test for the reported results.

        Returns
        -------
        bool: True if both schema and metadata match the truth source, else False.
        """
        targets = "_metadata vs truth"
        self.print_if_verbose(f"\t{targets}")
        passed = self.metadata_ds.schema.equals(
            self.constructed_truth_schema, check_metadata=self.args.check_metadata
        )
        self.results.append(
            Result(
                passed=passed, target=targets, test=test_info["test"], description=test_info["description"]
            )
        )
        return passed

    def _test_schema_file_footers(self, test_info: dict) -> bool:
        """Test the file footers schema and metadata against the truth schema.

        Parameters
        ----------
        test_info : dict
            Information about this test for the reported results.

        Returns
        -------
        bool: True if all schema and metadata tests pass, else False.
        """
        targets = "file footers vs truth"
        self.print_if_verbose(f"\t{targets}")

        bad_files = []
        for frag in self.files_ds.get_fragments():
            if not frag.physical_schema.equals(
                self.constructed_truth_schema, check_metadata=self.args.check_metadata
            ):
                bad_files.append(frag.path)
        bad_files = self._relative_paths(bad_files)

        passed = len(bad_files) == 0
        self.results.append(
            Result(
                passed=passed,
                target=targets,
                bad_files=bad_files,
                test=test_info["test"],
                description=test_info["description"],
            )
        )
        return passed

    def _relative_paths(self, absolute_paths):
        """Find the relative path for dataset parquet files,
        assuming a pattern like <base_path>/Norder=d/Dir=d/Npix=d"""
        relative_path_pattern = re.compile(r".*(Norder.*)")
        relative_paths = [str(relative_path_pattern.match(file).group(1)) or file for file in absolute_paths]
        return relative_paths

    def write_results(self) -> None:
        """Write the verification results to file at `args.output_path` / `args.output_filename`."""
        self.args.output_file_path.parent.mkdir(exist_ok=True, parents=True)
        # Write provenance info
        with open(self.args.output_file_path, self.args.write_mode, encoding="utf8") as fout:
            fout.writelines(
                [
                    "# HATS verification results for\n",
                    f"# {self.args.input_catalog_path}\n",
                    f"# Package versions: hats v{hats.__version__}; hats-import v{hats_import.__version__}\n",
                    f"# User-supplied truth schema: {self.args.truth_schema}\n",
                    f"# User-supplied truth total rows: {self.args.truth_total_rows}\n",
                ]
            )
        # Write results
        self.results_df.to_csv(self.args.output_file_path, mode="a", header=True, index=False)
        self.print_if_verbose(f"\nVerifier results written to {self.args.output_file_path}")

    def print_if_verbose(self, message):
        """If the args.verbose=True flag is enabled, print to standard out. Otherwise, no operation."""
        if self.args.verbose:
            print(message)



import pandas as pd
import pyarrow as pa
from hats.io import file_io
from pyarrow import csv

from hats_import.catalog.file_readers.input_reader import InputReader


class CsvReader(InputReader):
    """CSV reader for the most common CSV reading arguments.

    This uses `pandas.read_csv`, and you can find more information on
    additional arguments in the pandas documentation:
    https://pandas.pydata.org/docs/reference/api/pandas.read_csv.html

    Attributes:
        chunksize (int): number of rows to read in a single iteration.
        header (int, list of int, None, default 'infer'): rows to
            use as the header with column names
        schema_file (str): path to a parquet schema file. if provided, header names
            and column types will be pulled from the parquet schema metadata.
        column_names (list[str]): the names of columns if no header is available
        type_map (dict): the data types to use for columns
        parquet_kwargs (dict): additional keyword arguments to use when
            reading the parquet schema metadata, passed to pandas.read_parquet.
            See https://pandas.pydata.org/docs/reference/api/pandas.read_parquet.html
        kwargs (dict): additional keyword arguments to use when reading
            the CSV files with pandas.read_csv.
            See https://pandas.pydata.org/docs/reference/api/pandas.read_csv.html
    """

    def __init__(
        self,
        chunksize=500_000,
        header="infer",
        schema_file=None,
        column_names=None,
        type_map=None,
        parquet_kwargs=None,
        upath_kwargs=None,
        **kwargs,
    ):
        self.chunksize = chunksize
        self.header = header
        self.schema_file = schema_file
        self.column_names = column_names
        self.type_map = type_map
        self.parquet_kwargs = parquet_kwargs
        self.upath_kwargs = upath_kwargs
        self.kwargs = kwargs

        schema_parquet = None
        if self.schema_file:
            if self.parquet_kwargs is None:
                self.parquet_kwargs = {}
            schema_parquet = file_io.read_parquet_file_to_pandas(
                self.schema_file,
                **self.parquet_kwargs,
            )

        if self.column_names:
            self.kwargs["names"] = self.column_names
        elif not self.header and schema_parquet is not None:
            self.kwargs["names"] = list(schema_parquet.columns)

        if self.type_map:
            self.kwargs["dtype"] = self.type_map
        elif schema_parquet is not None:
            self.kwargs["dtype"] = schema_parquet.dtypes.to_dict()

    def read(self, input_file, read_columns=None):
        input_file = self.regular_file_exists(input_file, **self.kwargs)

        if read_columns:
            self.kwargs["usecols"] = read_columns

        return file_io.load_csv_to_pandas_generator(
            input_file,
            chunksize=self.chunksize,
            header=self.header,
            **self.kwargs,
        )


class CsvPyarrowReader(InputReader):
    """CSV reader that uses the pyarrow library for reading.

    This *can* be faster than the pandas reader, and *can* have different handling of
    data types.

    Attributes:
        chunksize (int): number of BYTES of the file to process at once.
            This is different from chunksize seen in other readers!!
            For large files, this can prevent loading the entire file
            into memory at once.
        column_names (list[str] or None): Names of columns to use from the input dataset.
            If None, use all columns.
        schema_file (str): path to a parquet schema file. if provided, column names and types
            will match those of the schema.
        read_options (csv.ReadOptions): options for reading CSV files using pyarrow.
            We will set the ``block_size`` argument based on the value for ``chunksize``.
            See https://arrow.apache.org/docs/python/generated/pyarrow.csv.ReadOptions.html
        convert_options (csv.ConvertOptions): options for converting CSV data to pyarrow Table.
            We will pass the pyarrow schema from ``schema_file`` to the ``column_types`` property.
            See https://arrow.apache.org/docs/python/generated/pyarrow.csv.ConvertOptions.html
        kwargs: arguments to pass along to pyarrow.parquet.ParquetFile.
            See https://arrow.apache.org/docs/python/generated/pyarrow.parquet.ParquetFile.html
    """

    def __init__(
        self,
        *,
        chunksize=10 * 1024**2,
        compression=None,
        column_names=None,
        schema_file=None,
        read_options=None,
        convert_options=None,
        **kwargs,
    ):
        self.kwargs = kwargs
        self.column_names = column_names
        self.read_options = read_options or csv.ReadOptions(block_size=chunksize)
        self.convert_options = convert_options or csv.ConvertOptions()
        self.compression = compression
        if schema_file:
            schema = file_io.read_parquet_metadata(schema_file).schema.to_arrow_schema()
            if not self.convert_options.column_types:
                self.convert_options.column_types = schema
            if not self.read_options.column_names:
                self.read_options.column_names = schema.names
                # Set skip_rows here to skip parsing the header row with column names.
                # See https://arrow.apache.org/docs/python/generated/pyarrow.csv.ReadOptions.html
                self.read_options.skip_rows = 1

    def read(self, input_file, read_columns=None):
        input_file = self.regular_file_exists(input_file, **self.kwargs)

        # If we only want to read some columns (e.g. radec), we must specify
        # in the convert_options argument
        read_columns = read_columns or self.column_names
        convert_options = self.convert_options
        if read_columns:
            convert_options.include_columns = read_columns
        with input_file.open(mode="rb", compression=self.compression) as file_handle:
            with csv.open_csv(
                file_handle, convert_options=convert_options, read_options=self.read_options, **self.kwargs
            ) as reader:
                for next_chunk in reader:
                    table = pa.Table.from_batches([next_chunk])
                    table = table.replace_schema_metadata()
                    yield table


class IndexedCsvReader(CsvReader):
    """Reads an index file, containing paths to CSV files to be read and batched

    See CsvReader for additional configuration for reading CSV files.
    """

    def read(self, input_file, read_columns=None):
        file_paths = self.read_index_file(
            input_file=input_file, upath_kwargs=self.upath_kwargs, **self.kwargs
        )

        batch_size = 0
        batch_frames = []
        for file in file_paths:
            for single_frame in super().read(file, read_columns=read_columns):
                if batch_size + len(single_frame) >= self.chunksize:
                    # We've hit our chunksize, send the batch off to the task.
                    if len(batch_frames) == 0:
                        yield single_frame
                        batch_size = 0
                    else:
                        yield pd.concat(batch_frames, ignore_index=True)
                        batch_frames = []
                        batch_frames.append(single_frame)
                        batch_size = len(single_frame)
                else:
                    batch_frames.append(single_frame)
                    batch_size += len(single_frame)

        if len(batch_frames) > 0:
            yield pd.concat(batch_frames, ignore_index=True)



import hats.io.file_io as io

import hats_import.catalog.run_import as catalog_runner
import hats_import.index.run_index as index_runner
import hats_import.margin_cache.margin_cache as margin_runner
from hats_import.collection.arguments import CollectionArguments
from hats_import.pipeline_resume_plan import print_progress


def run(args, client):
    """Run catalog collection creation pipeline."""
    if not args:
        raise TypeError("args is required and should be type CollectionArguments")
    if not isinstance(args, CollectionArguments):
        raise TypeError("args must be type CollectionArguments")

    catalog_args = args.get_catalog_args()

    if catalog_args:
        catalog_runner.run(catalog_args, client)

    for margin_args in args.get_margin_args():
        margin_runner.generate_margin_cache(margin_args, client)

    for index_args in args.get_index_args():
        index_runner.run(index_args, client)

    ## Finishing collection
    with print_progress(
        total=2,
        stage_name="Finishing",
        use_progress_bar=args.progress_bar,
        simple_progress_bar=args.simple_progress_bar,
    ) as step_progress:
        collection_info = args.to_collection_properties()
        collection_info.to_properties_file(args.catalog_path)
        step_progress.update(1)
        io.remove_directory(args.tmp_path, ignore_errors=True)
        step_progress.update(1)



"""Tests of argument validation"""

import pytest

from hats_import.catalog.arguments import ImportArguments, check_healpix_order_range

# pylint: disable=protected-access


def test_none():
    """No arguments provided. Should error for required args."""
    with pytest.raises(ValueError):
        ImportArguments()


def test_empty_required(small_sky_parts_dir, tmp_path):
    """*Most* required arguments are provided."""
    ## File reader is missing
    with pytest.raises(ValueError, match="file_reader"):
        ImportArguments(
            output_artifact_name="catalog",
            file_reader=None,
            input_path=small_sky_parts_dir,
            output_path=tmp_path,
        )
    ## Input path is missing
    with pytest.raises(ValueError, match="input_file"):
        ImportArguments(
            output_artifact_name="catalog",
            file_reader="csv",
            input_path="",
            output_path=tmp_path,
        )

    with pytest.raises(ValueError, match="input_file_list is required"):
        ImportArguments(
            output_artifact_name="catalog",
            file_reader="csv",
            input_file_list=None,
            output_path=tmp_path,
        )

    with pytest.raises(ValueError, match="input_file_list is empty"):
        ImportArguments(
            output_artifact_name="catalog",
            file_reader="csv",
            input_file_list=[],
            output_path=tmp_path,
        )


def test_invalid_paths(blank_data_dir, tmp_path):
    """Required arguments are provided, but paths aren't found."""
    ## Prove that it works with required args
    ImportArguments(
        output_artifact_name="catalog",
        input_path=blank_data_dir,
        file_reader="csv",
        output_path=tmp_path,
        progress_bar=False,
    )

    ## Bad input path
    with pytest.raises(FileNotFoundError):
        ImportArguments(
            output_artifact_name="catalog",
            input_path="path",
            file_reader="csv",
            output_path=tmp_path,
        )

    # Too many paths!
    with pytest.raises(ValueError, match="exactly one of"):
        ImportArguments(
            output_artifact_name="catalog",
            input_path=blank_data_dir,
            input_file_list=[blank_data_dir],
            file_reader="csv",
            output_path=tmp_path,
            progress_bar=False,
        )


def test_missing_paths(tmp_path):
    ## Input path has no files
    with pytest.raises(FileNotFoundError):
        ImportArguments(
            output_artifact_name="catalog",
            file_reader="csv",
            input_path=tmp_path,
            output_path=tmp_path,
        )


def test_good_paths(blank_data_dir, blank_data_file, tmp_path):
    """Required arguments are provided, and paths are found."""
    tmp_path_str = str(tmp_path)
    args = ImportArguments(
        output_artifact_name="catalog",
        input_path=blank_data_dir,
        file_reader="csv",
        output_path=tmp_path_str,
        tmp_dir=tmp_path_str,
        progress_bar=False,
    )
    assert args.input_path == blank_data_dir
    assert len(args.input_paths) == 1
    assert str(blank_data_file) in str(args.input_paths[0])


def test_multiple_files_in_path(small_sky_parts_dir, tmp_path):
    """Required arguments are provided, and paths are found."""
    args = ImportArguments(
        output_artifact_name="catalog",
        input_path=small_sky_parts_dir,
        file_reader="csv",
        output_path=tmp_path,
        progress_bar=False,
    )
    assert args.input_path == small_sky_parts_dir
    assert len(args.input_paths) == 6


def test_single_debug_file(formats_headers_csv, tmp_path):
    """Required arguments are provided, and paths are found."""
    args = ImportArguments(
        output_artifact_name="catalog",
        input_file_list=[formats_headers_csv],
        file_reader="csv",
        output_path=tmp_path,
        progress_bar=False,
    )
    assert len(args.input_paths) == 1
    assert args.input_paths[0] == formats_headers_csv

    args = ImportArguments(
        output_artifact_name="catalog",
        input_file_list=formats_headers_csv,
        file_reader="csv",
        output_path=tmp_path,
        progress_bar=False,
    )
    assert len(args.input_paths) == 1
    assert args.input_paths[0] == formats_headers_csv

    single_file_name = formats_headers_csv.name
    assert isinstance(single_file_name, str)
    args = ImportArguments(
        output_artifact_name="catalog",
        input_file_list=single_file_name,
        file_reader="csv",
        output_path=tmp_path,
        progress_bar=False,
    )
    assert len(args.input_paths) == 1
    assert args.input_paths[0] == single_file_name


def test_healpix_args(blank_data_dir, tmp_path):
    """Test errors for healpix partitioning arguments"""
    with pytest.raises(ValueError, match="highest_healpix_order"):
        ImportArguments(
            output_artifact_name="catalog",
            input_path=blank_data_dir,
            file_reader="csv",
            output_path=tmp_path,
            highest_healpix_order=30,
        )
    with pytest.raises(ValueError, match="pixel_threshold"):
        ImportArguments(
            output_artifact_name="catalog",
            input_path=blank_data_dir,
            file_reader="csv",
            output_path=tmp_path,
            pixel_threshold=3,
        )
    with pytest.raises(ValueError, match="constant_healpix_order"):
        ImportArguments(
            output_artifact_name="catalog",
            input_path=blank_data_dir,
            file_reader="csv",
            output_path=tmp_path,
            constant_healpix_order=30,
        )


def test_catalog_type(blank_data_dir, tmp_path):
    """Test errors for catalog types."""
    with pytest.raises(ValueError, match="catalog_type"):
        ImportArguments(
            output_artifact_name="catalog",
            catalog_type=None,
            input_path=blank_data_dir,
            file_reader="csv",
            output_path=tmp_path,
        )

    with pytest.raises(ValueError, match="catalog_type"):
        ImportArguments(
            output_artifact_name="catalog",
            catalog_type="association",
            input_path=blank_data_dir,
            file_reader="csv",
            output_path=tmp_path,
        )


def test_use_healpix_29(blank_data_dir, tmp_path):
    with pytest.raises(ValueError, match="no sort columns should be added"):
        ImportArguments(
            output_artifact_name="catalog",
            input_path=blank_data_dir,
            file_reader="csv",
            output_path=tmp_path,
            use_healpix_29=True,
            sort_columns="foo",
        )
    ImportArguments(
        output_artifact_name="catalog",
        input_path=blank_data_dir,
        file_reader="csv",
        output_path=tmp_path,
        use_healpix_29=True,
        sort_columns="",  # empty string is ok
    )


def test_to_table_properties(blank_data_dir, tmp_path):
    """Verify creation of catalog parameters for catalog to be created."""
    args = ImportArguments(
        output_artifact_name="catalog",
        input_path=blank_data_dir,
        file_reader="csv",
        output_path=tmp_path,
        tmp_dir=tmp_path,
        progress_bar=False,
        addl_hats_properties={"hats_cols_default": "id, mjd", "obs_regime": "Optical"},
    )
    catalog_info = args.to_table_properties(
        total_rows=10, highest_order=4, moc_sky_fraction=22 / 7, column_names=["id", "mjd", "ra", "dec"]
    )
    assert catalog_info.catalog_name == "catalog"
    assert catalog_info.total_rows == 10
    assert catalog_info.default_columns == ["id", "mjd"]
    assert catalog_info.__pydantic_extra__["obs_regime"] == "Optical"


def test_to_table_properties_missing_default(blank_data_dir, tmp_path):
    """Verify creation of catalog parameters for catalog to be created."""
    args = ImportArguments(
        output_artifact_name="catalog",
        input_path=blank_data_dir,
        file_reader="csv",
        output_path=tmp_path,
        tmp_dir=tmp_path,
        progress_bar=False,
        addl_hats_properties={"hats_cols_default": "id, mjd", "obs_regime": "Optical"},
    )
    with pytest.raises(ValueError, match="Some default columns not found"):
        args.to_table_properties(
            total_rows=10, highest_order=4, moc_sky_fraction=22 / 7, column_names=["id", "ra", "dec"]
        )


def test_check_healpix_order_range():
    """Test method check_healpix_order_range"""
    check_healpix_order_range(5, "order_field")
    check_healpix_order_range(5, "order_field", lower_bound=0, upper_bound=19)

    with pytest.raises(ValueError, match="positive"):
        check_healpix_order_range(5, "order_field", lower_bound=-1)

    with pytest.raises(ValueError, match="29"):
        check_healpix_order_range(5, "order_field", upper_bound=30)

    with pytest.raises(ValueError, match="order_field"):
        check_healpix_order_range(-1, "order_field")
    with pytest.raises(ValueError, match="order_field"):
        check_healpix_order_range(30, "order_field")

    with pytest.raises(TypeError, match="not supported"):
        check_healpix_order_range("two", "order_field")
    with pytest.raises(TypeError, match="not supported"):
        check_healpix_order_range(5, "order_field", upper_bound="ten")


def test_no_import_overwrite(small_sky_object_catalog, parquet_shards_dir):
    """Runner should refuse to overwrite a valid catalog"""
    catalog_dir = small_sky_object_catalog.parent
    catalog_name = small_sky_object_catalog.name
    with pytest.raises(ValueError, match="already contains a valid catalog"):
        ImportArguments(
            input_path=parquet_shards_dir,
            output_path=catalog_dir,
            output_artifact_name=catalog_name,
            file_reader="parquet",
        )



"""Utility to hold all arguments required throughout verification pipeline"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Literal

import hats.io.paths
from hats.io import file_io
from upath import UPath


@dataclass(kw_only=True)
class VerificationArguments:
    """Container for verification arguments."""

    input_catalog_path: UPath = field()
    """Path to an existing catalog that will be inspected. This must be a directory
    containing (at least) the hats ancillary files and a 'dataset/' directory
    containing the parquet dataset. Can be supplied as a string or path object."""
    output_path: UPath = field()
    """Directory where the verification report should be written.
     Can be supplied as a string or path object."""
    output_filename: str = field(default="verifier_results.csv")
    """Filename for the verification report."""
    truth_total_rows: int | None = field(default=None)
    """Total number of rows expected in this catalog."""
    truth_schema: UPath | None = field(default=None)
    """Path to a parquet file or dataset containing the expected schema. If None (default),
    the catalog's _common_metadata file will be used. This schema will be used to verify
    all non-hats columns and (optionally) the file-level metadata. Can be supplied as a
    string or path object."""
    verbose: bool = True
    """Should we output progress and results to standard out?"""
    write_mode: Literal["a", "w", "x"] = "a"
    """Mode to be used when writing the output file. Options have the typical meanings:
        - 'w': truncate the file first
        - 'x': exclusive creation, failing if the file already exists
        - 'a': append to the end of file if it exists"""
    check_metadata: bool = False
    """Whether to check the metadata as well as the schema."""

    @property
    def input_dataset_path(self) -> UPath:
        """Path to the directory under `input_catalog_path` that contains the parquet dataset."""
        return file_io.append_paths_to_pointer(self.input_catalog_path, hats.io.paths.DATASET_DIR)

    @property
    def output_file_path(self) -> UPath:
        """Path to the output file (`output_path` / `output_filename`)."""
        return file_io.append_paths_to_pointer(self.output_path, self.output_filename)

    def __post_init__(self) -> None:
        self.input_catalog_path = file_io.get_upath(self.input_catalog_path)
        self.output_path = file_io.get_upath(self.output_path)

        if self.truth_schema is not None:
            self.truth_schema = file_io.append_paths_to_pointer(self.truth_schema)
            if not self.truth_schema.exists():
                raise FileNotFoundError("truth_schema must be an existing file or directory")



"""Test end-to-end execution of pipeline with different formats and configurations.

Please add a brief description in the docstring of the features or specific
regression the test case is exercising.
"""

# pylint: disable=too-many-lines
import glob
import os
from pathlib import Path

import numpy as np
import numpy.testing as npt
import pandas as pd
import pyarrow as pa
import pyarrow.dataset as pds
import pyarrow.parquet as pq
import pytest
from hats import read_hats
from hats.io.skymap import read_skymap
from hats.pixel_math import HealpixPixel
from hats.pixel_math.spatial_index import SPATIAL_INDEX_COLUMN, spatial_index_to_healpix
from pyarrow import csv

import hats_import.catalog.run_import as runner
from hats_import.catalog.arguments import ImportArguments
from hats_import.catalog.file_readers import CsvReader, get_file_reader
from hats_import.catalog.file_readers.csv import CsvPyarrowReader
from hats_import.catalog.file_readers.parquet import ParquetPyarrowReader


@pytest.mark.dask
def test_import_snappy_source_table(
    dask_client,
    small_sky_source_dir,
    tmp_path,
):
    """Test basic execution, using a larger source file.
    - catalog type should be source
    - will have larger partition info than the corresponding object catalog
    """
    args = ImportArguments(
        output_artifact_name="small_sky_source_catalog.parquet",
        input_path=small_sky_source_dir,
        file_reader="csv",
        catalog_type="source",
        ra_column="source_ra",
        dec_column="source_dec",
        sort_columns="source_id",
        output_path=tmp_path,
        dask_tmp=tmp_path,
        highest_healpix_order=2,
        pixel_threshold=3_000,
        progress_bar=False,
        # Sneak in a test of setting a non-default compression scheme.
        write_table_kwargs={"compression": "SNAPPY"},
    )

    runner.run(args, dask_client)

    # Check that the catalog metadata file exists
    catalog = read_hats(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert catalog.catalog_info.ra_column == "source_ra"
    assert catalog.catalog_info.dec_column == "source_dec"
    assert len(catalog.get_healpix_pixels()) == 14

    output_file = os.path.join(args.catalog_path, "dataset", "Norder=1", "Dir=0", "Npix=47.parquet")
    metadata = pq.read_metadata(output_file)
    assert metadata.num_row_groups == 1
    assert metadata.row_group(0).column(0).compression == "SNAPPY"

    # Check that the pixels have been sorted by the default _healpix_29 values
    pixel_data = pq.read_table(output_file)
    sorting_columns = metadata.row_group(0).sorting_columns
    ordering_tuples = pq.SortingColumn.to_ordering(pixel_data.schema, sorting_columns)[0]
    assert ordering_tuples[0] == (SPATIAL_INDEX_COLUMN, "ascending")
    assert pixel_data.equals(pixel_data.sort_by(SPATIAL_INDEX_COLUMN))


@pytest.mark.dask
def test_import_with_num_row_groups(
    dask_client,
    small_sky_source_dir,
    tmp_path,
):
    """The row group size will be specified in `row_group_kwargs`"""
    args = ImportArguments(
        output_artifact_name="small_sky_source_catalog",
        input_path=small_sky_source_dir,
        file_reader="csv",
        catalog_type="source",
        ra_column="source_ra",
        dec_column="source_dec",
        sort_columns="source_id",
        output_path=tmp_path,
        dask_tmp=tmp_path,
        highest_healpix_order=2,
        pixel_threshold=3_000,
        progress_bar=False,
        # Sneak in a test of custom row group size
        row_group_kwargs={"num_rows": 100},
    )

    runner.run(args, dask_client)

    # Check that the catalog metadata file exists
    catalog = read_hats(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert catalog.catalog_info.ra_column == "source_ra"
    assert catalog.catalog_info.dec_column == "source_dec"
    assert len(catalog.get_healpix_pixels()) == 14

    output_file = os.path.join(args.catalog_path, "dataset", "Norder=1", "Dir=0", "Npix=47.parquet")
    pixel = pq.ParquetFile(output_file)
    data = pixel.read()
    assert len(pixel.read()) == 2395

    # Check that the number of row groups is the one expected
    metadata = pixel.metadata
    assert metadata.num_rows == 2395
    # 24 = math.ceil(metadata.num_rows / 100)
    assert metadata.num_row_groups == 24
    # The last row group has fewer number of rows, which is fine
    assert all(metadata.row_group(i).num_rows == 100 for i in range(metadata.num_row_groups - 1))

    # Check that the sorting columns were saved in the parquet metadata
    sorting_columns = metadata.row_group(0).sorting_columns
    ordering_tuples = pq.SortingColumn.to_ordering(data.schema, sorting_columns)[0]
    assert ordering_tuples[0] == (SPATIAL_INDEX_COLUMN, "ascending")
    assert ordering_tuples[1] == ("source_id", "ascending")

    # And that the data is indeed sorted
    sorted_data = data.sort_by([(SPATIAL_INDEX_COLUMN, "ascending"), ("source_id", "ascending")])
    assert data.equals(sorted_data)


@pytest.mark.dask
def test_import_with_subtile_row_groups(
    dask_client,
    small_sky_source_dir,
    tmp_path,
):
    """The row group subtile splitting will be specified in `row_group_kwargs`"""
    args = ImportArguments(
        output_artifact_name="small_sky_source_catalog",
        input_path=small_sky_source_dir,
        file_reader="csv",
        catalog_type="source",
        ra_column="source_ra",
        dec_column="source_dec",
        sort_columns="source_id",
        output_path=tmp_path,
        dask_tmp=tmp_path,
        highest_healpix_order=2,
        pixel_threshold=3_000,
        progress_bar=False,
        # Sneak in a test of custom subtile splitting
        row_group_kwargs={"subtile_order_delta": 2},
    )

    runner.run(args, dask_client)

    # Check that the catalog metadata file exists
    catalog = read_hats(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert catalog.catalog_info.ra_column == "source_ra"
    assert catalog.catalog_info.dec_column == "source_dec"
    assert len(catalog.get_healpix_pixels()) == 14

    output_file = os.path.join(args.catalog_path, "dataset", "Norder=1", "Dir=0", "Npix=47.parquet")

    pixel = pq.ParquetFile(output_file)
    assert len(pixel.read()) == 2395

    # Check that the number of row groups is the one expected
    metadata = pixel.metadata
    assert metadata.num_rows == 2395
    child_pixels = HealpixPixel(1, 47).convert_to_higher_order(delta_order=2)
    # The empty sub-tiles are not kept
    assert pixel.num_row_groups <= len(child_pixels)

    seen_pixels = []

    for i in range(metadata.num_row_groups):
        row_group = metadata.row_group(i)

        # Check that the row group statistics are correct
        assert row_group.num_rows > 0
        min_healpix29 = row_group.column(0).statistics.min
        max_healpix29 = row_group.column(0).statistics.max
        pixel_min, pixel_max = spatial_index_to_healpix([min_healpix29, max_healpix29], target_order=3)
        assert pixel_min == pixel_max and HealpixPixel(3, pixel_min) in child_pixels

        # The row group contains data that does in fact belong to the pixel
        row_group_healpix29 = pixel.read_row_group(i)[SPATIAL_INDEX_COLUMN].to_numpy()
        assert all(row_group_healpix29 >= min_healpix29)
        assert all(row_group_healpix29 <= max_healpix29)
        seen_pixels.append(pixel_min)

    # Make sure there was no overlap between row group pixels
    assert list(set(seen_pixels)) == seen_pixels


@pytest.mark.dask
def test_import_mixed_schema_csv(
    dask_client,
    mixed_schema_csv_dir,
    mixed_schema_csv_parquet,
    assert_parquet_file_ids,
    tmp_path,
):
    """Test basic execution, with a mixed schema.
    - the two input files in `mixed_schema_csv_dir` have different *implied* schemas
        when parsed by pandas. this verifies that they end up with the same schema
        and can be combined into a single parquet file.
    - this additionally uses pathlib.Path for all path inputs.
    """
    args = ImportArguments(
        output_artifact_name="mixed_csv_bad",
        input_file_list=[
            Path(mixed_schema_csv_dir) / "input_01.csv",
            Path(mixed_schema_csv_dir) / "input_02.csv",
            Path(mixed_schema_csv_dir) / "input_03.csv",
        ],
        output_path=tmp_path,
        dask_tmp=tmp_path,
        highest_healpix_order=1,
        file_reader=get_file_reader(
            "csv",
            chunksize=1,
            schema_file=Path(mixed_schema_csv_parquet),
        ),
        progress_bar=False,
    )

    runner.run(args, dask_client)

    # Check that the catalog parquet file exists
    output_file = os.path.join(args.catalog_path, "dataset", "Norder=0", "Dir=0", "Npix=11.parquet")

    assert_parquet_file_ids(output_file, "id", [*range(700, 708)])

    # Check that the schema is correct for leaf parquet and _metadata files
    expected_parquet_schema = pa.schema(
        [
            pa.field("_healpix_29", pa.int64()),
            pa.field("id", pa.int64()),
            pa.field("ra", pa.float64()),
            pa.field("dec", pa.float64()),
            pa.field("ra_error", pa.int64()),
            pa.field("dec_error", pa.int64()),
            pa.field("comment", pa.string()),
            pa.field("code", pa.string()),
        ]
    )
    schema = pq.read_metadata(output_file).schema.to_arrow_schema()
    assert schema.equals(expected_parquet_schema)
    schema = pq.read_metadata(args.catalog_path / "dataset" / "_metadata").schema.to_arrow_schema()
    assert schema.equals(expected_parquet_schema)


@pytest.mark.dask
def test_import_mixed_schema_csv_pyarrow(
    dask_client,
    mixed_schema_csv_dir,
    mixed_schema_csv_parquet,
    assert_parquet_file_ids,
    tmp_path,
):
    """Test basic execution, with a mixed schema, reading with pyarrow's CSV reader.
    - the two input files in `mixed_schema_csv_dir` have different *implied* schemas
        when parsed by pandas. this verifies that they end up with the same schema
        and can be combined into a single parquet file.
    - this additionally uses pathlib.Path for all path inputs.
    """
    input_files = list(Path(mixed_schema_csv_dir).glob("*.csv"))
    input_files.sort()
    args = ImportArguments(
        output_artifact_name="mixed_csv_bad",
        # Pyarrow CSV reader isn't happy about empty files.
        # See https://github.com/apache/arrow/discussions/46429
        input_file_list=input_files[0:2],
        output_path=tmp_path,
        dask_tmp=tmp_path,
        highest_healpix_order=1,
        file_reader=CsvPyarrowReader(
            schema_file=mixed_schema_csv_parquet,
            read_options=csv.ReadOptions(
                column_names=["id", "ra", "dec", "ra_error", "dec_error", "comment", "code"],
                skip_rows=1,
            ),
        ),
        progress_bar=False,
    )

    runner.run(args, dask_client)

    # Check that the catalog parquet file exists
    output_file = os.path.join(args.catalog_path, "dataset", "Norder=0", "Dir=0", "Npix=11.parquet")

    assert_parquet_file_ids(output_file, "id", [*range(700, 708)])

    # Check that the schema is correct for leaf parquet and _metadata files
    expected_parquet_schema = pa.schema(
        [
            pa.field("_healpix_29", pa.int64()),
            pa.field("id", pa.int64()),
            pa.field("ra", pa.float64()),
            pa.field("dec", pa.float64()),
            pa.field("ra_error", pa.int64()),
            pa.field("dec_error", pa.int64()),
            pa.field("comment", pa.string()),
            pa.field("code", pa.string()),
        ]
    )
    schema = pq.read_metadata(output_file).schema.to_arrow_schema()
    assert schema.equals(expected_parquet_schema)
    schema = pq.read_metadata(args.catalog_path / "dataset" / "_metadata").schema.to_arrow_schema()
    assert schema.equals(expected_parquet_schema)


@pytest.mark.dask
def test_import_preserve_index(
    dask_client,
    formats_pandasindex,
    assert_parquet_file_ids,
    assert_parquet_file_index,
    tmp_path,
):
    """Test basic execution, with input with pandas metadata.
    - the input file is a parquet file with some pandas metadata.
        this verifies that the parquet file at the end also still has the
        contents of that column.
    """

    expected_indexes = [
        "star1_1",
        "star1_2",
        "star1_3",
        "star1_4",
        "galaxy1_1",
        "galaxy1_2",
        "galaxy2_1",
        "galaxy2_2",
    ]
    assert_parquet_file_index(formats_pandasindex, expected_indexes)
    data_frame = pd.read_parquet(formats_pandasindex, engine="pyarrow")
    assert data_frame.index.name == "obs_id"
    npt.assert_array_equal(
        data_frame.columns,
        ["obj_id", "band", "ra", "dec", "mag"],
    )

    ## Don't generate a hats spatial index. Verify that the original index remains.
    args = ImportArguments(
        output_artifact_name="pandasindex",
        input_file_list=[formats_pandasindex],
        file_reader="parquet",
        sort_columns="obs_id",
        add_healpix_29=False,
        output_path=tmp_path,
        dask_tmp=tmp_path,
        highest_healpix_order=1,
        progress_bar=False,
    )

    runner.run(args, dask_client)

    # Check that the catalog parquet file exists
    output_file = os.path.join(args.catalog_path, "dataset", "Norder=1", "Dir=0", "Npix=46.parquet")

    data_frame = pd.read_parquet(output_file, engine="pyarrow")
    npt.assert_array_equal(
        data_frame.columns,
        ["obs_id", "obj_id", "band", "ra", "dec", "mag"],
    )

    ## DO generate a hats spatial index. Verify that the original index is preserved in a column.
    args = ImportArguments(
        output_artifact_name="pandasindex_preserve",
        input_file_list=[formats_pandasindex],
        file_reader="parquet",
        sort_columns="obs_id",
        add_healpix_29=True,
        output_path=tmp_path,
        dask_tmp=tmp_path,
        highest_healpix_order=1,
        progress_bar=False,
    )

    runner.run(args, dask_client)

    # Check that the catalog parquet file exists
    output_file = os.path.join(args.catalog_path, "dataset", "Norder=1", "Dir=0", "Npix=46.parquet")

    data_frame = pd.read_parquet(output_file, engine="pyarrow")
    npt.assert_array_equal(
        data_frame.columns,
        ["_healpix_29", "obs_id", "obj_id", "band", "ra", "dec", "mag"],
    )
    assert_parquet_file_ids(output_file, "obs_id", expected_indexes)


@pytest.mark.dask
def test_import_constant_healpix_order(
    dask_client,
    small_sky_parts_dir,
    tmp_path,
):
    """Test basic execution.
    - tests that all the final tiles are at the same healpix order,
        and that we don't create tiles where there is no data.
    """
    args = ImportArguments(
        output_artifact_name="small_sky_object_catalog",
        input_path=small_sky_parts_dir,
        file_reader="csv",
        output_path=tmp_path,
        dask_tmp=tmp_path,
        constant_healpix_order=2,
        progress_bar=False,
    )

    runner.run(args, dask_client)

    # Check that the catalog metadata file exists
    catalog = read_hats(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    # Check that the partition info file exists - all pixels at order 2!
    assert all(pixel.order == 2 for pixel in catalog.partition_info.get_healpix_pixels())

    # Pick a parquet file and make sure it contains as many rows as we expect
    output_file = os.path.join(args.catalog_path, "dataset", "Norder=2", "Dir=0", "Npix=178.parquet")

    data_frame = pd.read_parquet(output_file, engine="pyarrow")
    assert len(data_frame) == 14
    ids = data_frame["id"]
    assert np.logical_and(ids >= 700, ids < 832).all()

    assert catalog.catalog_info.skymap_alt_orders is None
    assert catalog.catalog_info.skymap_order == 2


@pytest.mark.dask
def test_import_keep_intermediate_files(
    dask_client,
    small_sky_parts_dir,
    tmp_path,
):
    """Test that ALL intermediate files are still around on-disk after
    successful import, when setting the appropriate flags.
    """
    temp = tmp_path / "intermediate_files"
    temp.mkdir(parents=True)
    args = ImportArguments(
        output_artifact_name="small_sky_object_catalog",
        input_path=small_sky_parts_dir,
        file_reader="csv",
        output_path=tmp_path,
        tmp_dir=temp,
        dask_tmp=temp,
        progress_bar=False,
        highest_healpix_order=2,
        delete_intermediate_parquet_files=False,
        delete_resume_log_files=False,
    )
    runner.run(args, dask_client)

    # Check that the catalog metadata file exists
    catalog = read_hats(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path

    # Check that both stage level and intermediate parquet files exist
    base_intermediate_dir = temp / "small_sky_object_catalog" / "intermediate"
    assert_stage_level_files_exist(base_intermediate_dir)
    assert_intermediate_parquet_files_exist(base_intermediate_dir)


@pytest.mark.dask
def test_import_delete_provided_temp_directory(
    dask_client,
    small_sky_parts_dir,
    tmp_path_factory,
):
    """Test that ALL intermediate files (and temporary base directory) are deleted
    after successful import, when both `delete_intermediate_parquet_files` and
    `delete_resume_log_files` are set to True."""
    output_dir = tmp_path_factory.mktemp("catalogs")
    # Provided temporary directory, outside `output_dir`
    temp = tmp_path_factory.mktemp("intermediate_files")

    # When at least one of the delete flags is set to False we do
    # not delete the provided temporary base directory.
    args = ImportArguments(
        output_artifact_name="keep_log_files",
        input_path=small_sky_parts_dir,
        file_reader="csv",
        output_path=output_dir,
        tmp_path=temp,
        dask_tmp=temp,
        progress_bar=False,
        highest_healpix_order=2,
        delete_intermediate_parquet_files=True,
        delete_resume_log_files=False,
    )
    runner.run(args, dask_client)
    assert_stage_level_files_exist(temp / "keep_log_files" / "intermediate")

    args = ImportArguments(
        output_artifact_name="keep_parquet_intermediate",
        input_path=small_sky_parts_dir,
        file_reader="csv",
        output_path=output_dir,
        tmp_path=temp,
        dask_tmp=temp,
        progress_bar=False,
        highest_healpix_order=2,
        delete_intermediate_parquet_files=False,
        delete_resume_log_files=True,
        resume=False,
    )
    runner.run(args, dask_client)
    assert_intermediate_parquet_files_exist(temp / "keep_parquet_intermediate" / "intermediate")

    # The temporary directory is deleted.
    args = ImportArguments(
        output_artifact_name="remove_all_intermediate",
        input_path=small_sky_parts_dir,
        file_reader="csv",
        output_path=output_dir,
        tmp_path=temp,
        dask_tmp=temp,
        progress_bar=False,
        highest_healpix_order=2,
        delete_intermediate_parquet_files=True,
        delete_resume_log_files=True,
        resume=False,
    )
    runner.run(args, dask_client)
    assert not os.path.exists(temp / "remove_all_intermediate")


def assert_stage_level_files_exist(base_intermediate_dir):
    # Check that stage-level done files are still around for the import of
    # `small_sky_object_catalog` at order 0.
    expected_contents = [
        "alignment.pickle",
        "histograms",  # directory containing sub-histograms
        "input_paths.txt",  # original input paths for subsequent comparison
        "mapping_done",  # stage-level done file
        "mapping_histogram.npz",  # concatenated histogram file
        "order_0",  # all intermediate parquet files
        "reader.pickle",  # pickled InputReader
        "reducing",  # directory containing task-level done files
        "reducing_done",  # stage-level done file
        "splitting",  # directory containing task-level done files
        "splitting_done",  # stage-level done file
    ]
    assert_directory_contains(base_intermediate_dir, expected_contents)

    checking_dir = base_intermediate_dir / "histograms"
    assert_directory_contains(
        checking_dir, ["map_0.npz", "map_1.npz", "map_2.npz", "map_3.npz", "map_4.npz", "map_5.npz"]
    )
    checking_dir = base_intermediate_dir / "splitting"
    assert_directory_contains(
        checking_dir,
        ["split_0_done", "split_1_done", "split_2_done", "split_3_done", "split_4_done", "split_5_done"],
    )

    checking_dir = base_intermediate_dir / "reducing"
    assert_directory_contains(checking_dir, ["0_11_done"])


def assert_intermediate_parquet_files_exist(base_intermediate_dir):
    # Check that all the intermediate parquet shards are still around for the
    # import of `small_sky_object_catalog` at order 0.
    checking_dir = base_intermediate_dir / "order_0" / "dir_0" / "pixel_11"
    assert_directory_contains(
        checking_dir,
        [
            "shard_split_0_0.parquet",
            "shard_split_1_0.parquet",
            "shard_split_2_0.parquet",
            "shard_split_3_0.parquet",
            "shard_split_4_0.parquet",
        ],
    )


def assert_directory_contains(dir_name, expected_contents):
    assert os.path.exists(dir_name)
    actual_contents = os.listdir(dir_name)
    actual_contents.sort()
    npt.assert_array_equal(actual_contents, expected_contents)


@pytest.mark.dask
def test_import_lowest_healpix_order(
    dask_client,
    small_sky_parts_dir,
    tmp_path,
):
    """Test basic execution.
    - tests that all the final tiles are at the lowest healpix order,
        and that we don't create tiles where there is no data.
    """
    args = ImportArguments(
        output_artifact_name="small_sky_object_catalog",
        input_path=small_sky_parts_dir,
        file_reader="csv",
        output_path=tmp_path,
        dask_tmp=tmp_path,
        lowest_healpix_order=2,
        highest_healpix_order=4,
        skymap_alt_orders=[3, 1],
        progress_bar=False,
    )

    runner.run(args, dask_client)

    # Check that the catalog metadata file exists
    catalog = read_hats(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    # Check that the partition info file exists - all pixels at order 2!
    assert all(pixel.order >= 2 for pixel in catalog.get_healpix_pixels())
    assert len(catalog.get_healpix_pixels()) == 14

    # Pick a parquet file and make sure it contains as many rows as we expect
    output_file = os.path.join(args.catalog_path, "dataset", "Norder=2", "Dir=0", "Npix=178.parquet")

    data_frame = pd.read_parquet(output_file, engine="pyarrow")
    assert len(data_frame) == 14
    ids = data_frame["id"]
    assert np.logical_and(ids >= 700, ids < 832).all()

    point_map_file = args.catalog_path / "point_map.fits"
    assert point_map_file.exists()

    skymap_file = args.catalog_path / "skymap.fits"
    assert skymap_file.exists()

    assert len(read_skymap(catalog, None)) == 3_072

    skymap_file = args.catalog_path / "skymap.1.fits"
    assert skymap_file.exists()
    skymap_file = args.catalog_path / "skymap.2.fits"
    assert not skymap_file.exists()
    skymap_file = args.catalog_path / "skymap.3.fits"
    assert skymap_file.exists()

    assert catalog.catalog_info.skymap_alt_orders == [1, 3]
    assert catalog.catalog_info.skymap_order == 4

    old_properties_file = args.catalog_path / "properties"
    assert old_properties_file.exists()
    new_properties_file = args.catalog_path / "hats.properties"
    assert new_properties_file.exists()

    with open(old_properties_file, "r", encoding="utf-8") as old_file:
        old_contents = old_file.readlines()

    with open(new_properties_file, "r", encoding="utf-8") as new_file:
        new_contents = new_file.readlines()

    assert old_contents == new_contents


class StarrReader(CsvReader):
    """Shallow subclass"""

    def read(self, input_file, read_columns=None):
        files = glob.glob(f"{input_file}/*.starr")
        files.sort()
        for file in files:
            return super().read(file, read_columns)


@pytest.mark.dask
def test_import_starr_file(
    dask_client,
    formats_dir,
    assert_parquet_file_ids,
    tmp_path,
):
    """Test basic execution.
    - tests that we can run pipeline with a totally unknown file type, so long
      as a valid InputReader implementation is provided.
    """

    args = ImportArguments(
        output_artifact_name="starr",
        input_file_list=[formats_dir],
        file_reader=StarrReader(),
        output_path=tmp_path,
        dask_tmp=tmp_path,
        highest_healpix_order=2,
        pixel_threshold=3_000,
        progress_bar=False,
    )

    runner.run(args, dask_client)

    # Check that the catalog metadata file exists
    catalog = read_hats(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert catalog.catalog_info.total_rows == 131
    assert len(catalog.get_healpix_pixels()) == 1

    # Check that the catalog parquet file exists and contains correct object IDs
    output_file = os.path.join(args.catalog_path, "dataset", "Norder=0", "Dir=0", "Npix=11.parquet")

    expected_ids = [*range(700, 831)]
    assert_parquet_file_ids(output_file, "id", expected_ids)


class PyarrowCsvReader(CsvReader):
    """Use pyarrow for CSV reading, and force some pyarrow dtypes.
    Return a pyarrow table instead of pd.DataFrame."""

    def read(self, input_file, read_columns=None):
        table = csv.read_csv(input_file)
        extras = pa.array([[True, False, True]] * len(table), type=pa.list_(pa.bool_(), 3))
        table = table.append_column("extras", extras)
        yield table


@pytest.mark.dask
def test_import_pyarrow_types(
    dask_client,
    small_sky_single_file,
    assert_parquet_file_ids,
    tmp_path,
):
    """Test basic execution.
    - tests that we can run pipeline with a totally unknown file type, so long
      as a valid InputReader implementation is provided.
    """

    args = ImportArguments(
        output_artifact_name="pyarrow_dtype",
        input_file_list=[small_sky_single_file],
        file_reader=PyarrowCsvReader(),
        output_path=tmp_path,
        dask_tmp=tmp_path,
        highest_healpix_order=2,
        pixel_threshold=3_000,
        progress_bar=False,
    )

    runner.run(args, dask_client)

    # Check that the catalog metadata file exists
    catalog = read_hats(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert catalog.catalog_info.total_rows == 131
    assert len(catalog.get_healpix_pixels()) == 1

    # Check that the catalog parquet file exists and contains correct object IDs
    output_file = args.catalog_path / "dataset" / "Norder=0" / "Dir=0" / "Npix=11.parquet"

    expected_ids = [*range(700, 831)]
    assert_parquet_file_ids(output_file, "id", expected_ids)

    expected_parquet_schema = pa.schema(
        [
            pa.field("_healpix_29", pa.int64()),
            pa.field("id", pa.int64()),
            pa.field("ra", pa.float64()),
            pa.field("dec", pa.float64()),
            pa.field("ra_error", pa.int64()),
            pa.field("dec_error", pa.int64()),
            pa.field("extras", pa.list_(pa.bool_(), 3)),  # The 3 is the length for `fixed_size_list`
        ]
    )
    schema = pq.read_metadata(output_file).schema.to_arrow_schema()
    assert schema.equals(expected_parquet_schema)
    schema = pq.read_metadata(args.catalog_path / "dataset" / "_metadata").schema.to_arrow_schema()
    assert schema.equals(expected_parquet_schema)


@pytest.mark.dask
def test_import_healpix_29_pyarrow_table_csv(
    dask_client,
    small_sky_single_file,
    assert_parquet_file_ids,
    tmp_path,
):
    """Should be identical to the above test, but uses the CsvPyarrowReader."""
    args = ImportArguments(
        output_artifact_name="small_sky_pyarrow",
        input_file_list=[small_sky_single_file],
        file_reader=CsvPyarrowReader(),
        output_path=tmp_path,
        dask_tmp=tmp_path,
        highest_healpix_order=2,
        pixel_threshold=3_000,
        progress_bar=False,
    )

    runner.run(args, dask_client)

    # Check that the catalog metadata file exists
    catalog = read_hats(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert catalog.catalog_info.total_rows == 131
    assert len(catalog.get_healpix_pixels()) == 1

    # Check that the catalog parquet file exists and contains correct object IDs
    output_file = args.catalog_path / "dataset" / "Norder=0" / "Dir=0" / "Npix=11.parquet"

    expected_ids = [*range(700, 831)]
    assert_parquet_file_ids(output_file, "id", expected_ids)
    data_frame = pd.read_parquet(output_file, engine="pyarrow")
    assert data_frame.index.name is None
    npt.assert_array_equal(
        data_frame.columns,
        ["_healpix_29", "id", "ra", "dec", "ra_error", "dec_error"],
    )


@pytest.mark.dask
def test_import_healpix_29_pyarrow_table_parquet(
    dask_client,
    formats_dir,
    assert_parquet_file_ids,
    tmp_path,
):
    """Should be identical to the above test, but uses the ParquetPyarrowReader."""
    input_file = formats_dir / "healpix_29_index.parquet"
    args = ImportArguments(
        output_artifact_name="using_healpix_index",
        input_file_list=[input_file],
        file_reader=ParquetPyarrowReader(),
        output_path=tmp_path,
        dask_tmp=tmp_path,
        use_healpix_29=True,
        highest_healpix_order=2,
        pixel_threshold=3_000,
        progress_bar=False,
        ## Sneak in a test for custom extension
        npix_suffix=".pq",
    )

    runner.run(args, dask_client)

    # Check that the catalog metadata file exists
    catalog = read_hats(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert catalog.catalog_info.total_rows == 131
    assert len(catalog.get_healpix_pixels()) == 1

    # Check that the catalog parquet file exists and contains correct object IDs
    output_file = args.catalog_path / "dataset" / "Norder=0" / "Dir=0" / "Npix=11.pq"

    expected_ids = [*range(700, 831)]
    assert_parquet_file_ids(output_file, "id", expected_ids)
    data_frame = pd.read_parquet(output_file, engine="pyarrow")

    npt.assert_array_equal(
        data_frame.columns,
        ["id", "_healpix_29"],
    )


@pytest.mark.dask
def test_import_healpix_29(
    dask_client,
    formats_dir,
    assert_parquet_file_ids,
    tmp_path,
):
    """Test basic execution, using a previously-computed _healpix_29 column for spatial partitioning."""
    ## First, let's just check the assumptions we have about our input file:
    ## - should have _healpix_29 as the indexed column
    ## - should NOT have any columns like "ra" or "dec"
    input_file = formats_dir / "healpix_29_index.parquet"

    expected_ids = [*range(700, 831)]
    assert_parquet_file_ids(input_file, "id", expected_ids)

    data_frame = pd.read_parquet(input_file, engine="pyarrow")
    assert data_frame.index.name == "_healpix_29"
    npt.assert_array_equal(data_frame.columns, ["id"])

    args = ImportArguments(
        output_artifact_name="using_healpix_29",
        input_file_list=[input_file],
        file_reader="parquet",
        output_path=tmp_path,
        dask_tmp=tmp_path,
        use_healpix_29=True,
        highest_healpix_order=2,
        pixel_threshold=3_000,
        progress_bar=False,
    )

    runner.run(args, dask_client)

    # Check that the catalog metadata file exists
    catalog = read_hats(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert catalog.catalog_info.total_rows == 131
    assert len(catalog.get_healpix_pixels()) == 1

    # Check that the catalog parquet file exists and contains correct object IDs
    output_file = os.path.join(args.catalog_path, "dataset", "Norder=0", "Dir=0", "Npix=11.parquet")

    expected_ids = [*range(700, 831)]
    assert_parquet_file_ids(output_file, "id", expected_ids)
    data_frame = pd.read_parquet(output_file, engine="pyarrow")
    npt.assert_array_equal(
        data_frame.columns,
        ["_healpix_29", "id"],
    )


@pytest.mark.dask
def test_import_healpix_29_no_pandas(
    dask_client,
    formats_dir,
    assert_parquet_file_ids,
    tmp_path,
):
    """Test basic execution, using a previously-computed _healpix_29 column for spatial partitioning."""
    input_file = formats_dir / "spatial_index.csv"
    args = ImportArguments(
        output_artifact_name="using_healpix_29",
        input_file_list=[input_file],
        file_reader="csv",
        output_path=tmp_path,
        dask_tmp=tmp_path,
        use_healpix_29=True,
        highest_healpix_order=2,
        pixel_threshold=3_000,
        progress_bar=False,
    )

    runner.run(args, dask_client)

    # Check that the catalog metadata file exists
    catalog = read_hats(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert catalog.catalog_info.total_rows == 131
    assert len(catalog.get_healpix_pixels()) == 1

    # Check that the catalog parquet file exists and contains correct object IDs
    output_file = os.path.join(args.catalog_path, "dataset", "Norder=0", "Dir=0", "Npix=11.parquet")

    expected_ids = [*range(700, 831)]
    assert_parquet_file_ids(output_file, "id", expected_ids)
    data_frame = pd.read_parquet(output_file, engine="pyarrow")
    npt.assert_array_equal(
        data_frame.columns,
        ["id", "_healpix_29", "magnitude", "nobs"],
    )


@pytest.mark.dask
def test_import_gaia_minimum(
    dask_client,
    formats_dir,
    tmp_path,
):
    """Test end-to-end import, using a representative chunk of gaia data."""
    input_file = formats_dir / "gaia_minimum.csv"
    schema_file = formats_dir / "gaia_minimum_schema.parquet"

    args = ImportArguments(
        output_artifact_name="gaia_minimum",
        input_file_list=[input_file],
        file_reader=CsvReader(
            comment="#",
            schema_file=schema_file,
        ),
        ra_column="ra",
        dec_column="dec",
        sort_columns="solution_id",
        use_schema_file=schema_file,
        output_path=tmp_path,
        dask_tmp=tmp_path,
        highest_healpix_order=2,
        pixel_threshold=3_000,
        progress_bar=False,
    )

    runner.run(args, dask_client)

    # Check that the catalog metadata file exists
    catalog = read_hats(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert catalog.catalog_info.total_rows == 5
    assert len(catalog.get_healpix_pixels()) == 3

    # Pick an output file, and make sure it has valid columns:
    first_pixel = catalog.get_healpix_pixels()[0]
    output_file = (
        args.catalog_path
        / "dataset"
        / f"Norder={first_pixel.order}"
        / "Dir=0"
        / f"Npix={first_pixel.pixel}.parquet"
    )
    data_frame = pd.read_parquet(output_file)

    # Make sure that the spatial index values match the pixel for the partition (0,5)
    spatial_index_pixels = spatial_index_to_healpix(data_frame["_healpix_29"].values, 0)
    npt.assert_array_equal(spatial_index_pixels, [5, 5, 5])


@pytest.mark.dask
def test_import_issue_538(
    dask_client,
    reproducers_dir,
    tmp_path,
):
    """Test end-to-end import, using a reproducible chunk of data."""
    args = ImportArguments(
        output_artifact_name="numpy_unique",
        input_file_list=[reproducers_dir / "issue_538.csv"],
        file_reader="csv",
        ra_column="RA",
        dec_column="DEC",
        output_path=tmp_path,
        dask_tmp=tmp_path,
        highest_healpix_order=2,
        pixel_threshold=3_000,
        progress_bar=False,
    )

    runner.run(args, dask_client)

    # Check that the catalog metadata file exists
    catalog = read_hats(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert catalog.catalog_info.total_rows == 132
    assert len(catalog.get_healpix_pixels()) == 12


@pytest.mark.dask
def test_gaia_ecsv(
    dask_client,
    formats_dir,
    tmp_path,
    assert_parquet_file_ids,
):
    input_file = formats_dir / "gaia_epoch.ecsv"

    args = ImportArguments(
        output_artifact_name="gaia_e_astropy",
        input_file_list=[input_file],
        file_reader="ecsv",
        ra_column="ra",
        dec_column="dec",
        sort_columns="solution_id,source_id",
        output_path=tmp_path,
        dask_tmp=tmp_path,
        highest_healpix_order=2,
        pixel_threshold=3_000,
        progress_bar=False,
    )

    runner.run(args, dask_client)

    # Check that the catalog metadata file exists
    catalog = read_hats(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert catalog.catalog_info.total_rows == 3
    assert len(catalog.get_healpix_pixels()) == 1

    first_pixel = catalog.get_healpix_pixels()[0]
    output_file = (
        args.catalog_path
        / "dataset"
        / f"Norder={first_pixel.order}"
        / "Dir=0"
        / f"Npix={first_pixel.pixel}.parquet"
    )
    assert_parquet_file_ids(output_file, "source_id", [10655814178816, 10892037246720, 14263587225600])

    # Check that the schema is correct for leaf parquet and _metadata files
    expected_parquet_schema = pa.schema(
        [
            pa.field("_healpix_29", pa.int64()),
            pa.field("solution_id", pa.int64()),
            pa.field("source_id", pa.int64()),
            pa.field("ra", pa.float64()),
            pa.field("dec", pa.float64()),
            pa.field("n_transits", pa.int16()),
            pa.field("transit_id", pa.list_(pa.int64())),
            pa.field("g_transit_time", pa.list_(pa.float64())),
            pa.field("g_transit_flux", pa.list_(pa.float64())),
            pa.field("g_transit_flux_error", pa.list_(pa.float64())),
            pa.field("g_transit_flux_over_error", pa.list_(pa.float32())),
            pa.field("g_transit_mag", pa.list_(pa.float64())),
            pa.field("g_transit_n_obs", pa.list_(pa.int8())),
            pa.field("bp_obs_time", pa.list_(pa.float64())),
            pa.field("bp_flux", pa.list_(pa.float64())),
            pa.field("bp_flux_error", pa.list_(pa.float64())),
            pa.field("bp_flux_over_error", pa.list_(pa.float32())),
            pa.field("bp_mag", pa.list_(pa.float64())),
            pa.field("rp_obs_time", pa.list_(pa.float64())),
            pa.field("rp_flux", pa.list_(pa.float64())),
            pa.field("rp_flux_error", pa.list_(pa.float64())),
            pa.field("rp_flux_over_error", pa.list_(pa.float32())),
            pa.field("rp_mag", pa.list_(pa.float64())),
            pa.field("photometry_flag_noisy_data", pa.list_(pa.bool_())),
            pa.field("photometry_flag_sm_unavailable", pa.list_(pa.bool_())),
            pa.field("photometry_flag_af1_unavailable", pa.list_(pa.bool_())),
            pa.field("photometry_flag_af2_unavailable", pa.list_(pa.bool_())),
            pa.field("photometry_flag_af3_unavailable", pa.list_(pa.bool_())),
            pa.field("photometry_flag_af4_unavailable", pa.list_(pa.bool_())),
            pa.field("photometry_flag_af5_unavailable", pa.list_(pa.bool_())),
            pa.field("photometry_flag_af6_unavailable", pa.list_(pa.bool_())),
            pa.field("photometry_flag_af7_unavailable", pa.list_(pa.bool_())),
            pa.field("photometry_flag_af8_unavailable", pa.list_(pa.bool_())),
            pa.field("photometry_flag_af9_unavailable", pa.list_(pa.bool_())),
            pa.field("photometry_flag_bp_unavailable", pa.list_(pa.bool_())),
            pa.field("photometry_flag_rp_unavailable", pa.list_(pa.bool_())),
            pa.field("photometry_flag_sm_reject", pa.list_(pa.bool_())),
            pa.field("photometry_flag_af1_reject", pa.list_(pa.bool_())),
            pa.field("photometry_flag_af2_reject", pa.list_(pa.bool_())),
            pa.field("photometry_flag_af3_reject", pa.list_(pa.bool_())),
            pa.field("photometry_flag_af4_reject", pa.list_(pa.bool_())),
            pa.field("photometry_flag_af5_reject", pa.list_(pa.bool_())),
            pa.field("photometry_flag_af6_reject", pa.list_(pa.bool_())),
            pa.field("photometry_flag_af7_reject", pa.list_(pa.bool_())),
            pa.field("photometry_flag_af8_reject", pa.list_(pa.bool_())),
            pa.field("photometry_flag_af9_reject", pa.list_(pa.bool_())),
            pa.field("photometry_flag_bp_reject", pa.list_(pa.bool_())),
            pa.field("photometry_flag_rp_reject", pa.list_(pa.bool_())),
            pa.field("variability_flag_g_reject", pa.list_(pa.bool_())),
            pa.field("variability_flag_bp_reject", pa.list_(pa.bool_())),
            pa.field("variability_flag_rp_reject", pa.list_(pa.bool_())),
        ]
    )

    # In-memory schema uses list<item> naming convention, but pyarrow converts to
    # the parquet-compliant list<element> convention when writing to disk.
    # Round trip the schema to get a schema with compliant nested naming convention.
    schema_path = tmp_path / "temp_schema.parquet"
    pq.write_table(expected_parquet_schema.empty_table(), where=schema_path)
    expected_parquet_schema = pq.read_metadata(schema_path).schema.to_arrow_schema()

    schema = pq.read_metadata(output_file).schema.to_arrow_schema()
    assert schema.equals(expected_parquet_schema)
    schema = pq.read_metadata(args.catalog_path / "dataset" / "_metadata").schema.to_arrow_schema()
    assert schema.equals(expected_parquet_schema)
    schema = pq.read_metadata(args.catalog_path / "dataset" / "_common_metadata").schema.to_arrow_schema()
    assert schema.equals(expected_parquet_schema)
    schema = pds.dataset(args.catalog_path, format="parquet").schema
    assert schema.equals(expected_parquet_schema)


@pytest.mark.dask
def test_import_indexed_csv(
    dask_client,
    indexed_files_dir,
    tmp_path,
):
    """Use indexed-style CSV reads. There are two index files, and we expect
    to have two batches worth of intermediate files."""
    temp = tmp_path / "intermediate_files"
    os.makedirs(temp)

    args = ImportArguments(
        output_artifact_name="indexed_csv",
        input_file_list=[
            indexed_files_dir / "csv_list_double_1_of_2.txt",
            indexed_files_dir / "csv_list_double_2_of_2.txt",
            indexed_files_dir / "csv_list_empty.txt",
        ],
        output_path=tmp_path,
        file_reader="indexed_csv",
        sort_columns="id",
        tmp_dir=temp,
        dask_tmp=temp,
        highest_healpix_order=2,
        delete_intermediate_parquet_files=False,
        delete_resume_log_files=False,
        pixel_threshold=3_000,
        progress_bar=False,
    )

    runner.run(args, dask_client)

    # Check that the catalog metadata file exists
    catalog = read_hats(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert len(catalog.get_healpix_pixels()) == 1

    # Check that there are TWO intermediate parquet file (two index files).
    assert_directory_contains(
        temp / "indexed_csv" / "intermediate" / "order_0" / "dir_0" / "pixel_11",
        [
            "shard_split_0_0.parquet",
            "shard_split_1_0.parquet",
        ],
    )


@pytest.mark.dask
def test_import_healpix13(dask_client, formats_dir, tmp_path):
    """Use catalog data with an existing spatial index NOT at order 29."""

    args = ImportArguments(
        output_artifact_name="healpix13",
        input_file_list=formats_dir / "small_sky_healpix13.csv",
        output_path=tmp_path,
        file_reader="csv",
        sort_columns="id",
        pixel_threshold=3_000,
        highest_healpix_order=2,
        progress_bar=False,
        add_healpix_29=False,
        addl_hats_properties={"hats_col_healpix": "healpix13", "hats_col_healpix_order": 13},
    )

    runner.run(args, dask_client)

    # Check that the catalog metadata file exists
    catalog = read_hats(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert len(catalog.get_healpix_pixels()) == 1
    assert catalog.has_healpix_column()
    assert catalog.catalog_info.healpix_column == "healpix13"


@pytest.mark.dask
def test_pickled_reader_class_issue542(
    dask_client,
    formats_dir,
    tmp_path,
):
    """Check if we can use reader class not imported from somewhere else,
    e.g. defined in a Jupyter notebook.
    """

    class MyReader(StarrReader):
        """My cool reader."""

    args = ImportArguments(
        output_artifact_name="starr",
        input_file_list=[formats_dir],
        file_reader=MyReader(),
        output_path=tmp_path,
        dask_tmp=tmp_path,
        highest_healpix_order=2,
        pixel_threshold=3_000,
        progress_bar=False,
    )

    runner.run(args, dask_client)

    # Check that the catalog metadata file exists
    catalog = read_hats(args.catalog_path)
    assert catalog.catalog_info.total_rows == 131



import pyarrow as pa
from hats.io import file_io

from hats_import.catalog.file_readers.input_reader import InputReader


class ParquetReader(InputReader):
    """Parquet reader for the most common Parquet reading arguments.

    Attributes:
        chunksize (int): number of rows of the file to process at once.
            For large files, this can prevent loading the entire file
            into memory at once.
        column_names (list[str] or None): Names of columns to use from the input dataset.
            If None, use all columns.
        kwargs: arguments to pass along to pyarrow.parquet.ParquetFile.
            See https://arrow.apache.org/docs/python/generated/pyarrow.parquet.ParquetFile.html
    """

    def __init__(self, chunksize=500_000, column_names=None, **kwargs):
        self.chunksize = chunksize
        self.column_names = column_names
        self.kwargs = kwargs

    def read(self, input_file, read_columns=None):
        input_file = self.regular_file_exists(input_file, **self.kwargs)
        columns = read_columns or self.column_names
        parquet_file = file_io.read_parquet_file(input_file, **self.kwargs)
        for smaller_table in parquet_file.iter_batches(
            batch_size=self.chunksize, columns=columns, use_pandas_metadata=True
        ):
            yield smaller_table.to_pandas()


class ParquetPyarrowReader(InputReader):
    """Parquet reader for the most common Parquet reading arguments.

    Attributes:
        chunksize (int): number of rows of the file to process at once.
            For large files, this can prevent loading the entire file
            into memory at once.
        column_names (list[str] or None): Names of columns to use from the input dataset.
            If None, use all columns.
        kwargs: arguments to pass along to pyarrow.parquet.ParquetFile.
            See https://arrow.apache.org/docs/python/generated/pyarrow.parquet.ParquetFile.html
    """

    def __init__(self, chunksize=500_000, column_names=None, **kwargs):
        self.chunksize = chunksize
        self.column_names = column_names
        self.kwargs = kwargs

    def read(self, input_file, read_columns=None):
        input_file = self.regular_file_exists(input_file, **self.kwargs)
        columns = read_columns or self.column_names
        parquet_file = file_io.read_parquet_file(input_file, **self.kwargs)
        for smaller_table in parquet_file.iter_batches(batch_size=self.chunksize, columns=columns):
            table = pa.Table.from_batches([smaller_table])
            table = table.replace_schema_metadata()
            yield table


class IndexedParquetReader(InputReader):
    """Reads an index file, containing paths to parquet files to be read and batched

    Attributes:
        chunksize (int): maximum number of rows to process at once.
            Large files will be processed in chunks. Small files will be concatenated.
            Also passed to pyarrow.dataset.Dataset.to_batches as `batch_size`.
        batch_readahead (int): number of batches to read ahead.
            Passed to pyarrow.dataset.Dataset.to_batches.
        fragment_readahead (int): number of fragments to read ahead.
            Passed to pyarrow.dataset.Dataset.to_batches.
        use_threads (bool): whether to use multiple threads for reading.
            Passed to pyarrow.dataset.Dataset.to_batches.
        column_names (list[str] or None): Names of columns to use from the input dataset.
            If None, use all columns.
        kwargs: additional arguments to pass along to InputReader.read_index_file.
    """

    def __init__(
        self,
        chunksize=500_000,
        batch_readahead=16,
        fragment_readahead=4,
        use_threads=True,
        column_names=None,
        upath_kwargs=None,
        **kwargs,
    ):
        self.chunksize = chunksize
        self.batch_readahead = batch_readahead
        self.fragment_readahead = fragment_readahead
        self.use_threads = use_threads
        self.column_names = column_names
        self.upath_kwargs = upath_kwargs
        self.kwargs = kwargs

    def read(self, input_file, read_columns=None):
        columns = read_columns or self.column_names
        file_names = self.read_index_file(
            input_file=input_file, upath_kwargs=self.upath_kwargs, **self.kwargs
        )
        (_, input_dataset) = file_io.read_parquet_dataset(file_names, **self.kwargs)

        batches, nrows = [], 0
        for batch in input_dataset.to_batches(
            batch_size=self.chunksize,
            batch_readahead=self.batch_readahead,
            fragment_readahead=self.fragment_readahead,
            use_threads=self.use_threads,
            columns=columns,
        ):
            if nrows + batch.num_rows > self.chunksize:
                # We've hit the chunksize so load to a DataFrame and yield.
                # There should always be at least one batch in here since batch_size == self.chunksize above.
                yield pa.Table.from_batches(batches).replace_schema_metadata()
                batches, nrows = [], 0

            batches.append(batch)
            nrows += batch.num_rows

        if len(batches) > 0:
            yield pa.Table.from_batches(batches).replace_schema_metadata()



import astropy.table
import numpy as np
import pyarrow as pa
from astropy.io import fits

from hats_import.catalog.file_readers.input_reader import InputReader


def _np_to_pyarrow_array(array: np.ndarray, *, flatten_tensors: bool) -> pa.Array:
    """Convert a numpy array to a pyarrow"""
    # We usually have the "wrong" byte order from FITS
    array = np.asanyarray(array, dtype=array.dtype.newbyteorder("="))
    values = pa.array(array.reshape(-1))
    # "Base" type
    if array.ndim == 1:
        return values
    # Flat multidimensional nested values if asked
    if flatten_tensors and array.ndim > 2:
        array = array.reshape(array.shape[0], -1)
    pa_list_array = pa.FixedSizeListArray.from_arrays(values, np.prod(array.shape[1:]))
    # An extra dimension is represented as a list array
    if array.ndim == 2:
        return pa_list_array
    # array.ndim > 2
    # Multiple extra dimensions are represented as a tensor array
    tensor_type = pa.fixed_shape_tensor(values.type, shape=array.shape[1:])
    return pa.FixedShapeTensorArray.from_storage(tensor_type, pa_list_array)


def _astropy_to_pyarrow_table(astropy_table: astropy.table.Table, *, flatten_tensors: bool) -> pa.Table:
    """Convert astropy.table.Table to pyarrow.Table"""
    pa_arrays = {}
    for column in astropy_table.columns:
        np_array = np.asarray(astropy_table[column])
        pa_arrays[column] = _np_to_pyarrow_array(np_array, flatten_tensors=flatten_tensors)
    return pa.table(pa_arrays)


def _first_table_hdu(hdul: fits.HDUList) -> int:
    """Get an index of the first HDU with a table"""
    for i, hdu in enumerate(hdul):
        if isinstance(hdu, (fits.TableHDU, fits.BinTableHDU, fits.GroupsHDU)):
            return i
    raise ValueError("No HDU with a table found")


class FitsReader(InputReader):
    """Chunked FITS file reader.

    There are two column-level arguments for reading fits files: ``column_names`` and ``skip_column_names``.

    - If neither is provided, we will read and process all columns in the fits file.
    - If ``column_names`` is given, we will use _only_ those names, and ``skip_column_names`` will be ignored.
    - If ``skip_column_names`` is provided, we will remove those columns from processing stages.

    NB: Uses astropy table memmap to avoid reading the entire file into memory.

    See: https://docs.astropy.org/en/stable/io/fits/index.html#working-with-large-files

    Attributes:
        chunksize (int): number of rows of the file to process at once.
            For large files, this can prevent loading the entire file
            into memory at once.
        column_names (list[str]): list of column names to keep. only use
            one of `column_names` or `skip_column_names`
        skip_column_names (list[str]): list of column names to skip. only use
            one of `column_names` or `skip_column_names`
        hdu (int | None): index of HDU to read. If None, use the first HDU
            with a table.
        flatten_tensors (bool): whether to flatten tensors. If True, the
            fixed-length list-array will be used, otherwise the arrow
            extension fixed-shape tensor will be used.
        fits_kwargs: keyword arguments passed along to ``astropy.io.fits.open(file_handler, **kwargs)``.
            See https://docs.astropy.org/en/stable/io/fits/api/files.html#astropy.io.fits.open
        table_kwargs: keyword arguments passed along to ``astropy.Table.read(hdu, **kwargs)``.
            See https://docs.astropy.org/en/stable/api/astropy.table.Table.html#astropy.table.Table.read
    """

    # First, we open FITS file "lazily": with the memory map and not decoding bytes to UTF8
    _default_fits_kwargs = {"memmap": True, "character_as_bytes": True}
    # Next, for each batch, we decode bytes
    _default_table_kwargs = {
        "format": "fits",
        "character_as_bytes": False,
    }

    def __init__(
        self,
        chunksize=500_000,
        column_names=None,
        skip_column_names=None,
        hdu: int | None = None,
        flatten_tensors: bool = True,
        fits_kwargs: dict[str, object] | None = None,
        table_kwargs: dict[str, object] | None = None,
    ):
        self.chunksize = chunksize
        self.column_names = column_names
        self.skip_column_names = None if skip_column_names is None else frozenset(skip_column_names)
        self.hdu_index = hdu
        self.flatten_tensors = flatten_tensors
        self.fits_kwargs = self._default_fits_kwargs | (fits_kwargs or {})
        self.table_kwargs = self._default_table_kwargs | (table_kwargs or {})

    def read(self, input_file, read_columns=None):
        input_file = self.regular_file_exists(input_file)
        with input_file.open("rb") as file_handle, fits.open(file_handle, **self.fits_kwargs) as hdul:
            if self.hdu_index is None:
                hdu_index = _first_table_hdu(hdul)
                hdu = hdul[hdu_index]
            else:
                hdu = hdul[self.hdu_index]

            column_names = hdu.columns.names
            if read_columns is not None:
                column_names = read_columns
            elif self.column_names is not None:
                column_names = self.column_names
            elif self.skip_column_names is not None:
                column_names = [col for col in column_names if col not in self.skip_column_names]

            for i_start in range(0, hdu.data.shape[0], self.chunksize):
                data_chunk = hdu.data[i_start : i_start + self.chunksize]
                hdu_chunk = type(hdu)(
                    data=data_chunk,
                    header=hdu.header,
                    ver=hdu.ver,
                )
                table_chunk = astropy.table.Table.read(hdu_chunk, **self.table_kwargs)
                yield _astropy_to_pyarrow_table(
                    table_chunk[column_names], flatten_tensors=self.flatten_tensors
                )



"""Tests of map reduce operations"""

import os
import shutil
from io import StringIO

import cloudpickle
import hats.pixel_math as hist
import hats.pixel_math.healpix_shim as hp
import numpy as np
import numpy.testing as npt
import pandas as pd
import pyarrow as pa
import pytest
from hats.pixel_math.sparse_histogram import SparseHistogram

import hats_import.catalog.map_reduce as mr
from hats_import.catalog.file_readers import get_file_reader
from hats_import.catalog.resume_plan import ResumePlan


def pickle_file_reader(tmp_path, file_reader) -> str:
    """Utility method to pickle a file reader, and return path to pickle."""
    pickled_reader_file = tmp_path / "reader.pickle"
    with open(pickled_reader_file, "wb") as pickle_file:
        cloudpickle.dump(file_reader, pickle_file)
    return pickled_reader_file


def test_read_empty_filename():
    """Empty file name"""
    with pytest.raises(FileNotFoundError):
        mr.map_to_pixels(
            input_file="",
            pickled_reader_file="",
            highest_order=10,
            ra_column="ra",
            dec_column="dec",
            resume_path="",
            mapping_key="map_0",
        )


def test_read_wrong_fileformat(small_sky_file0, tmp_path):
    """CSV file attempting to be read as parquet"""
    with pytest.raises(pa.lib.ArrowInvalid):
        mr.map_to_pixels(
            input_file=small_sky_file0,
            pickled_reader_file=pickle_file_reader(tmp_path, get_file_reader("parquet")),
            highest_order=0,
            ra_column="ra_mean",
            dec_column="dec_mean",
            resume_path="",
            mapping_key="map_0",
        )


def test_read_directory(test_data_dir, tmp_path):
    """Provide directory, not file"""
    with pytest.raises(FileNotFoundError):
        mr.map_to_pixels(
            input_file=test_data_dir,
            pickled_reader_file=pickle_file_reader(tmp_path, get_file_reader("parquet")),
            highest_order=0,
            ra_column="ra",
            dec_column="dec",
            resume_path="",
            mapping_key="map_0",
        )


def test_read_bad_fileformat(blank_data_file, capsys, tmp_path):
    """Unsupported file format"""
    with pytest.raises(NotImplementedError):
        mr.map_to_pixels(
            input_file=blank_data_file,
            pickled_reader_file=pickle_file_reader(tmp_path, None),
            highest_order=0,
            ra_column="ra",
            dec_column="dec",
            resume_path="",
            mapping_key="map_0",
        )
    captured = capsys.readouterr()
    assert "No file reader implemented" in captured.out


def read_partial_histogram(tmp_path, mapping_key):
    """Helper to read in the former result of a map operation."""
    histogram_file = tmp_path / "histograms" / f"{mapping_key}.npz"
    hist = SparseHistogram.from_file(histogram_file)
    return hist.to_array()


def test_read_single_fits(tmp_path, formats_fits):
    """Success case - fits file that exists being read as fits"""
    (tmp_path / "histograms").mkdir(parents=True)
    mr.map_to_pixels(
        input_file=formats_fits,
        pickled_reader_file=pickle_file_reader(tmp_path, get_file_reader("fits")),
        highest_order=0,
        ra_column="ra",
        dec_column="dec",
        resume_path=tmp_path,
        mapping_key="map_0",
    )

    result = read_partial_histogram(tmp_path, "map_0")
    assert len(result) == 12
    expected = hist.empty_histogram(0)
    expected[11] = 131
    npt.assert_array_equal(result, expected)


def test_map_headers_wrong(formats_headers_csv, tmp_path):
    """Test loading the a file with non-default headers (without specifying right headers)"""
    with pytest.raises(ValueError, match="columns expected but not found"):
        mr.map_to_pixels(
            input_file=formats_headers_csv,
            pickled_reader_file=pickle_file_reader(tmp_path, get_file_reader("csv")),
            highest_order=0,
            ra_column="ra",
            dec_column="dec",
            resume_path="",
            mapping_key="map_0",
        )


def test_map_headers(tmp_path, formats_headers_csv):
    """Test loading the a file with non-default headers"""
    (tmp_path / "histograms").mkdir(parents=True)
    mr.map_to_pixels(
        input_file=formats_headers_csv,
        pickled_reader_file=pickle_file_reader(tmp_path, get_file_reader("csv")),
        highest_order=0,
        ra_column="ra_mean",
        dec_column="dec_mean",
        resume_path=tmp_path,
        mapping_key="map_0",
    )

    result = read_partial_histogram(tmp_path, "map_0")

    assert len(result) == 12

    expected = hist.empty_histogram(0)
    expected[11] = 8
    npt.assert_array_equal(result, expected)
    assert (result == expected).all()


def test_map_with_healpix_29(tmp_path, formats_dir, small_sky_single_file):
    (tmp_path / "histograms").mkdir(parents=True)
    input_file = formats_dir / "spatial_index.csv"
    mr.map_to_pixels(
        input_file=input_file,
        pickled_reader_file=pickle_file_reader(tmp_path, get_file_reader("csv")),
        highest_order=0,
        ra_column="NOPE",
        dec_column="NOPE",
        use_healpix_29=True,  # radec don't matter. just use existing index
        resume_path=tmp_path,
        mapping_key="map_0",
    )

    expected = hist.empty_histogram(0)
    expected[11] = 131

    result = read_partial_histogram(tmp_path, "map_0")
    npt.assert_array_equal(result, expected)

    with pytest.raises(ValueError, match="columns expected but not found"):
        mr.map_to_pixels(
            input_file=small_sky_single_file,
            pickled_reader_file=pickle_file_reader(tmp_path, get_file_reader("csv")),
            highest_order=0,
            ra_column="NOPE",
            dec_column="NOPE",
            use_healpix_29=True,  # no pre-existing index! expect failure.
            resume_path=tmp_path,
            mapping_key="map_0",
        )


def test_map_with_schema(tmp_path, mixed_schema_csv_dir, mixed_schema_csv_parquet):
    """Test loading the a file when using a parquet schema file for dtypes"""
    (tmp_path / "histograms").mkdir(parents=True)
    input_file = mixed_schema_csv_dir / "input_01.csv"
    mr.map_to_pixels(
        input_file=input_file,
        pickled_reader_file=pickle_file_reader(
            tmp_path,
            get_file_reader(
                "csv",
                schema_file=mixed_schema_csv_parquet,
            ),
        ),
        highest_order=0,
        ra_column="ra",
        dec_column="dec",
        resume_path=tmp_path,
        mapping_key="map_0",
    )

    result = read_partial_histogram(tmp_path, "map_0")

    assert len(result) == 12

    expected = hist.empty_histogram(0)
    expected[11] = 4
    npt.assert_array_equal(result, expected)
    assert (result == expected).all()


def test_map_small_sky_order0(tmp_path, small_sky_single_file):
    """Test loading the small sky catalog and partitioning each object into the same large bucket"""
    (tmp_path / "histograms").mkdir(parents=True)
    mr.map_to_pixels(
        input_file=small_sky_single_file,
        pickled_reader_file=pickle_file_reader(tmp_path, get_file_reader("csv")),
        highest_order=0,
        ra_column="ra",
        dec_column="dec",
        resume_path=tmp_path,
        mapping_key="map_0",
    )

    result = read_partial_histogram(tmp_path, "map_0")

    assert len(result) == 12

    expected = hist.empty_histogram(0)
    expected[11] = 131
    npt.assert_array_equal(result, expected)
    assert (result == expected).all()


def test_map_small_sky_part_order1(tmp_path, small_sky_file0):
    """
    Test loading a small portion of the small sky catalog and
    partitioning objects into four smaller buckets
    """
    (tmp_path / "histograms").mkdir(parents=True)
    mr.map_to_pixels(
        input_file=small_sky_file0,
        pickled_reader_file=pickle_file_reader(tmp_path, get_file_reader("csv")),
        highest_order=1,
        ra_column="ra",
        dec_column="dec",
        resume_path=tmp_path,
        mapping_key="map_0",
    )

    result = read_partial_histogram(tmp_path, "map_0")

    assert len(result) == 48

    expected = hist.empty_histogram(1)
    filled_pixels = [5, 7, 11, 2]
    expected[44:] = filled_pixels[:]
    npt.assert_array_equal(result, expected)
    assert (result == expected).all()


def test_split_pixels_bad_format(blank_data_file, tmp_path, capsys):
    """Test error behavior, e.g. when alignment file is missing."""
    alignment = np.full(12, None)
    alignment[11] = (0, 11, 131)
    with pytest.raises(FileNotFoundError):
        mr.split_pixels(
            input_file=blank_data_file,
            pickled_reader_file="",
            highest_order=0,
            ra_column="ra_mean",
            dec_column="dec_mean",
            splitting_key="0",
            cache_shard_path=tmp_path,
            resume_path=tmp_path,
            alignment_file="",
        )
    captured = capsys.readouterr()
    assert "No such file or directory" in captured.out


def test_split_pixels_headers(formats_headers_csv, assert_parquet_file_ids, tmp_path):
    """Test loading a file with non-default headers"""
    plan = ResumePlan(tmp_path=tmp_path, progress_bar=False, input_paths=["foo1"])
    raw_histogram = np.full(12, 0)
    raw_histogram[11] = 131
    alignment_file = plan.get_alignment_file(raw_histogram, -1, 0, 0, 1_000, False, 131)
    mr.split_pixels(
        input_file=formats_headers_csv,
        pickled_reader_file=pickle_file_reader(tmp_path, get_file_reader("csv")),
        highest_order=0,
        ra_column="ra_mean",
        dec_column="dec_mean",
        splitting_key="0",
        cache_shard_path=tmp_path,
        resume_path=tmp_path,
        alignment_file=alignment_file,
    )

    file_name = tmp_path / "order_0" / "dir_0" / "pixel_11" / "shard_0_0.parquet"
    expected_ids = [*range(700, 708)]
    assert_parquet_file_ids(file_name, "object_id", expected_ids)

    file_name = tmp_path / "order_0" / "dir_0" / "pixel_1" / "shard_0_0.parquet"
    assert not os.path.exists(file_name)


def test_reduce_idempotent(parquet_shards_dir, assert_parquet_file_ids, tmp_path):
    """Reduce the shards, then reduce them again - make sure no error is thrown the second time."""
    shutil.copytree(parquet_shards_dir, tmp_path / "shards")
    shards_copy = tmp_path / "shards"
    (tmp_path / "reducing").mkdir(parents=True)

    mr.reduce_pixel_shards(
        cache_shard_path=shards_copy,
        resume_path=tmp_path,
        reducing_key="0_11",
        destination_pixel_order=0,
        destination_pixel_number=11,
        destination_pixel_size=131,
        output_path=tmp_path,
        add_healpix_29=True,
        ra_column="ra",
        dec_column="dec",
        sort_columns="id",
        delete_input_files=False,
    )

    output_file = tmp_path / "dataset" / "Norder=0" / "Dir=0" / "Npix=11.parquet"

    expected_ids = [*range(700, 831)]
    assert_parquet_file_ids(output_file, "id", expected_ids)

    ## Confirm idempotency of reduce operation.
    mr.reduce_pixel_shards(
        cache_shard_path=shards_copy,
        resume_path=tmp_path,
        reducing_key="0_11",
        destination_pixel_order=0,
        destination_pixel_number=11,
        destination_pixel_size=131,
        output_path=tmp_path,
        add_healpix_29=True,
        ra_column="ra",
        dec_column="dec",
        sort_columns="id",
        delete_input_files=True,
    )

    output_file = tmp_path / "dataset" / "Norder=0" / "Dir=0" / "Npix=11.parquet"

    expected_ids = [*range(700, 831)]
    assert_parquet_file_ids(output_file, "id", expected_ids)

    ## It's even ok to try to delete the files again!
    mr.reduce_pixel_shards(
        cache_shard_path=shards_copy,
        resume_path=tmp_path,
        reducing_key="0_11",
        destination_pixel_order=0,
        destination_pixel_number=11,
        destination_pixel_size=131,
        output_path=tmp_path,
        add_healpix_29=True,
        ra_column="ra",
        dec_column="dec",
        sort_columns="id",
        delete_input_files=True,
    )

    output_file = tmp_path / "dataset" / "Norder=0" / "Dir=0" / "Npix=11.parquet"

    expected_ids = [*range(700, 831)]
    assert_parquet_file_ids(output_file, "id", expected_ids)


def test_reduce_order0(parquet_shards_dir, assert_parquet_file_ids, tmp_path):
    """Test reducing into one large pixel"""
    (tmp_path / "reducing").mkdir(parents=True)
    mr.reduce_pixel_shards(
        cache_shard_path=parquet_shards_dir,
        resume_path=tmp_path,
        reducing_key="0_11",
        destination_pixel_order=0,
        destination_pixel_number=11,
        destination_pixel_size=131,
        output_path=tmp_path,
        add_healpix_29=True,
        ra_column="ra",
        dec_column="dec",
        sort_columns="id",
        delete_input_files=False,
    )

    output_file = tmp_path / "dataset" / "Norder=0" / "Dir=0" / "Npix=11.parquet"

    expected_ids = [*range(700, 831)]
    assert_parquet_file_ids(output_file, "id", expected_ids)


def test_reduce_healpix_29(parquet_shards_dir, assert_parquet_file_ids, tmp_path):
    """Test reducing with or without a _healpix_29 field"""
    (tmp_path / "reducing").mkdir(parents=True)
    mr.reduce_pixel_shards(
        cache_shard_path=parquet_shards_dir,
        resume_path=tmp_path,
        reducing_key="0_11",
        destination_pixel_order=0,
        destination_pixel_number=11,
        destination_pixel_size=131,
        output_path=tmp_path,
        ra_column="ra",
        dec_column="dec",
        sort_columns="id",
        delete_input_files=False,
    )

    output_file = tmp_path / "dataset" / "Norder=0" / "Dir=0" / "Npix=11.parquet"

    expected_ids = [*range(700, 831)]
    assert_parquet_file_ids(output_file, "id", expected_ids)
    data_frame = pd.read_parquet(output_file, engine="pyarrow")
    npt.assert_array_equal(
        data_frame.columns,
        ["_healpix_29", "id", "ra", "dec", "ra_error", "dec_error"],
    )

    output_file.unlink()

    mr.reduce_pixel_shards(
        cache_shard_path=parquet_shards_dir,
        resume_path=tmp_path,
        reducing_key="0_11",
        destination_pixel_order=0,
        destination_pixel_number=11,
        destination_pixel_size=131,
        output_path=tmp_path,
        add_healpix_29=False,  ## different from above
        ra_column="ra",
        dec_column="dec",
        sort_columns="id",
        delete_input_files=False,
    )

    assert_parquet_file_ids(output_file, "id", expected_ids)
    data_frame = pd.read_parquet(output_file, engine="pyarrow")
    ## No index name.
    assert data_frame.index.name is None
    ## Data fields are the same.
    npt.assert_array_equal(
        data_frame.columns,
        ["id", "ra", "dec", "ra_error", "dec_error"],
    )


def test_reduce_different_expectation(parquet_shards_dir, tmp_path):
    """Test reducing into one large pixel.

    This will succeed the first time, but we change the size expectation the second time,
    and the original file is in error."""
    (tmp_path / "reducing").mkdir(parents=True)
    mr.reduce_pixel_shards(
        cache_shard_path=parquet_shards_dir,
        resume_path=tmp_path,
        reducing_key="0_11",
        destination_pixel_order=0,
        destination_pixel_number=11,
        destination_pixel_size=131,
        output_path=tmp_path,
        ra_column="ra",
        dec_column="dec",
        sort_columns="id",
        delete_input_files=False,
    )
    with pytest.raises(ValueError, match="Unexpected number of objects in RESUMED"):
        mr.reduce_pixel_shards(
            cache_shard_path=parquet_shards_dir,
            resume_path=tmp_path,
            reducing_key="0_11",
            destination_pixel_order=0,
            destination_pixel_number=11,
            destination_pixel_size=11,  ## should be 131
            output_path=tmp_path,
            ra_column="ra",
            dec_column="dec",
            sort_columns="id",
            delete_input_files=False,
        )


def test_reduce_resume_corrupted(parquet_shards_dir, tmp_path):
    """Test reducing into one large pixel.

    This will succeed the first time, but we change the size expectation the second time,
    and the original file is in error."""
    (tmp_path / "reducing").mkdir(parents=True)
    output_dir = tmp_path / "dataset" / "Norder=0" / "Dir=0"
    output_dir.mkdir(parents=True)
    with open(output_dir / "Npix=11.parquet", "w", encoding="utf-8") as file_handle:
        file_handle.write("Not a good parquet file")
    with pytest.raises(pa.lib.ArrowInvalid, match="the file is corrupted or this is not a parquet file"):
        mr.reduce_pixel_shards(
            cache_shard_path=parquet_shards_dir,
            resume_path=tmp_path,
            reducing_key="0_11",
            destination_pixel_order=0,
            destination_pixel_number=11,
            destination_pixel_size=131,
            output_path=tmp_path,
            ra_column="ra",
            dec_column="dec",
            sort_columns="id",
            delete_input_files=False,
        )


def test_reduce_bad_expectation(parquet_shards_dir, tmp_path):
    """Test reducing into one large pixel"""
    with pytest.raises(ValueError, match="Unexpected number of objects"):
        mr.reduce_pixel_shards(
            cache_shard_path=parquet_shards_dir,
            resume_path=tmp_path,
            reducing_key="0_11",
            destination_pixel_order=0,
            destination_pixel_number=11,
            destination_pixel_size=11,  ## should be 131
            output_path=tmp_path,
            ra_column="ra",
            dec_column="dec",
            sort_columns="id",
            delete_input_files=False,
        )


def test_reduce_with_sorting_complex(assert_parquet_file_ids, tmp_path):
    """Test reducing and requesting specific sort columns.

    Logically, the input data has a mix of orderings in files, object IDs, and timestamps.
    Each source is partitioned according to the linked object's radec, and so will be
    ordered within the same spatial_index value.

    First, we take some time to set up these silly data points, then we test out
    reducing them into a single parquet file using a mix of reduction options.
    """
    (tmp_path / "reducing").mkdir(parents=True)
    shard_dir = tmp_path / "reduce_shards" / "order_0" / "dir_0" / "pixel_11"
    shard_dir.mkdir(parents=True)
    output_file = tmp_path / "dataset" / "Norder=0" / "Dir=0" / "Npix=11.parquet"

    file1_string = """source_id,object_id,time,ra,dec
1200,700,3000,282.5,-58.5
1201,700,4000,282.5,-58.5
1402,702,3000,310.5,-27.5
1403,702,3100,310.5,-27.5
1404,702,3200,310.5,-27.5
1505,703,4000,286.5,-69.5"""
    file1_data = pd.read_csv(StringIO(file1_string))
    file1_data.to_parquet(shard_dir / "file_1_shard_1.parquet")

    file2_string = """source_id,object_id,time,ra,dec
1206,700,2000,282.5,-58.5
1307,701,2200,299.5,-48.5
1308,701,2100,299.5,-48.5
1309,701,2000,299.5,-48.5"""
    file2_data = pd.read_csv(StringIO(file2_string))
    file2_data.to_parquet(shard_dir / "file_2_shard_1.parquet")

    combined_data = pd.concat([file1_data, file2_data])
    combined_data["norder19_healpix"] = hp.radec2pix(
        19,
        combined_data["ra"].values,
        combined_data["dec"].values,
    )
    ## Use this to prune generated columns like Norder, Npix, and _healpix_29
    comparison_columns = ["source_id", "object_id", "time", "ra", "dec"]

    ######################## Sort option 1: by source_id
    ## This will sort WITHIN an order 19 healpix pixel. In that ordering, the objects are
    ## (703, 700, 701, 702)
    mr.reduce_pixel_shards(
        cache_shard_path=tmp_path / "reduce_shards",
        resume_path=tmp_path,
        reducing_key="0_11",
        destination_pixel_order=0,
        destination_pixel_number=11,
        destination_pixel_size=10,
        output_path=tmp_path,
        ra_column="ra",
        dec_column="dec",
        sort_columns="source_id",
        delete_input_files=False,
    )

    ## sort order is effectively (norder19 healpix, source_id)
    data_frame = pd.read_parquet(output_file, engine="pyarrow")
    expected_dataframe = combined_data.sort_values(["norder19_healpix", "source_id"], kind="stable")
    pd.testing.assert_frame_equal(
        expected_dataframe[comparison_columns].reset_index(drop=True),
        data_frame[comparison_columns].reset_index(drop=True),
    )
    assert_parquet_file_ids(
        output_file,
        "source_id",
        [1505, 1200, 1201, 1206, 1307, 1308, 1309, 1402, 1403, 1404],
        resort_ids=False,
    )

    assert_parquet_file_ids(
        output_file,
        "object_id",
        [703, 700, 700, 700, 701, 701, 701, 702, 702, 702],
        resort_ids=False,
    )

    ######################## Sort option 2: by object id and time
    ## sort order is effectively (norder19 healpix, object id, time)
    output_file.unlink()
    mr.reduce_pixel_shards(
        cache_shard_path=tmp_path / "reduce_shards",
        resume_path=tmp_path,
        reducing_key="0_11",
        destination_pixel_order=0,
        destination_pixel_number=11,
        destination_pixel_size=10,
        output_path=tmp_path,
        ra_column="ra",
        dec_column="dec",
        sort_columns="object_id,time",
        delete_input_files=False,
    )

    data_frame = pd.read_parquet(output_file, engine="pyarrow")
    expected_dataframe = combined_data.sort_values(["norder19_healpix", "object_id", "time"], kind="stable")
    pd.testing.assert_frame_equal(
        expected_dataframe[comparison_columns].reset_index(drop=True),
        data_frame[comparison_columns].reset_index(drop=True),
    )
    assert_parquet_file_ids(
        output_file,
        "source_id",
        [1505, 1206, 1200, 1201, 1309, 1308, 1307, 1402, 1403, 1404],
        resort_ids=False,
    )
    assert_parquet_file_ids(
        output_file,
        "time",
        [4000, 2000, 3000, 4000, 2000, 2100, 2200, 3000, 3100, 3200],
        resort_ids=False,
    )

    ######################## Sort option 3: by object id and time WITHOUT spatial index.
    ## The 1500 block of ids goes back to the end, because we're not using
    ## spatial properties for sorting, only numeric.
    ## sort order is effectively (object id, time)
    output_file.unlink()
    mr.reduce_pixel_shards(
        cache_shard_path=tmp_path / "reduce_shards",
        resume_path=tmp_path,
        reducing_key="0_11",
        destination_pixel_order=0,
        destination_pixel_number=11,
        destination_pixel_size=10,
        output_path=tmp_path,
        ra_column="ra",
        dec_column="dec",
        sort_columns="object_id,time",
        add_healpix_29=False,
        delete_input_files=False,
    )

    data_frame = pd.read_parquet(output_file, engine="pyarrow")
    expected_dataframe = combined_data.sort_values(["object_id", "time"], kind="stable")
    pd.testing.assert_frame_equal(
        expected_dataframe[comparison_columns].reset_index(drop=True),
        data_frame[comparison_columns].reset_index(drop=True),
    )
    assert_parquet_file_ids(
        output_file,
        "source_id",
        [1206, 1200, 1201, 1309, 1308, 1307, 1402, 1403, 1404, 1505],
        resort_ids=False,
    )



"""Utility to hold all arguments required throughout hipscat -> hats conversion"""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

from hats.io import file_io
from upath import UPath

from hats_import.runtime_arguments import RuntimeArguments


@dataclass
class ConversionArguments(RuntimeArguments):
    """Data class for holding conversion arguments. Mostly just inheriting from RuntimeArguments"""

    ## Input
    input_catalog_path: str | Path | UPath | None = None

    def __post_init__(self):
        self._check_arguments()

    def _check_arguments(self):
        super()._check_arguments()
        if not self.input_catalog_path:
            raise ValueError("input_catalog_path is required")
        self.input_catalog_path = file_io.get_upath(self.input_catalog_path)
